__all__ = [
    "BasicLinter",
]


from beet import Context
from tokenstream import set_location

from mecha import AstCommand, AstSelector, Diagnostic, Mecha, Reducer, rule


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)
    mc.lint.extend(BasicLinter())


class BasicLinter(Reducer):
    """Linter with basic rules."""

    @rule(AstCommand, identifier="execute:subcommand")
    def execute_run(self, node: AstCommand):
        if isinstance(clause := node.arguments[0], AstCommand):
            if clause.identifier == "execute:run:subcommand":
                return set_location(
                    Diagnostic("warn", "Redundant `execute run` clause."),
                    node,
                    clause.arguments[0].location.with_horizontal_offset(-1),
                )

    @rule(AstCommand, identifier="execute:run:subcommand")
    def run_execute(self, node: AstCommand):
        if isinstance(clause := node.arguments[0], AstCommand):
            if clause.identifier == "execute:subcommand":
                return set_location(
                    Diagnostic("warn", "Redundant `run execute` clause."),
                    node,
                    clause.arguments[0].location.with_horizontal_offset(-1),
                )

    @rule(AstSelector)
    def selector_argument_order(self, node: AstSelector):
        order = [
            "type",
            "gamemode",
            "inverted gamemode",
            "team",
            "inverted team",
            "inverted type",
            "tag",
            "inverted tag",
            "name",
            "inverted name",
            "scores",
            "predicate",
            "inverted predicate",
            "advancements",
            "nbt",
        ]
        conflict = [-1] * len(order)

        for i, arg in enumerate(node.arguments):
            name = "inverted " * arg.inverted + arg.key.value

            try:
                index = order.index(name)
            except ValueError:
                continue

            j = conflict[index]

            if j >= 0:
                bad_arg = node.arguments[j]
                bad_arg_name = "inverted " * bad_arg.inverted + bad_arg.key.value

                d = Diagnostic(
                    level="warn",
                    message=f"{name.capitalize()} argument should go before {bad_arg_name}.",
                )

                yield set_location(d, arg)

            for conflict_index in range(index):
                if conflict[conflict_index] < 0:
                    conflict[conflict_index] = i



"""Plugin that resolves relative resource locations."""

__all__ = [
    "RelativeResourceLocationParser",
    "resolve_using_database",
    "resolve_relative_location",
]


from dataclasses import dataclass, replace
from pathlib import PurePosixPath
from typing import Tuple

from beet import Context
from tokenstream import InvalidSyntax, SourceLocation, TokenStream, set_location

from mecha import AstResourceLocation, CompilationDatabase, Mecha, Parser


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)
    mc.spec.parsers["resource_location_or_tag"] = RelativeResourceLocationParser(
        database=mc.database,
        parser=mc.spec.parsers["resource_location_or_tag"],
    )


@dataclass
class RelativeResourceLocationParser:
    """Parser that resolves relative resource locations."""

    database: CompilationDatabase
    parser: Parser

    def __call__(self, stream: TokenStream) -> AstResourceLocation:
        node: AstResourceLocation = self.parser(stream)

        if node.namespace is None and node.path.startswith(("./", "../")):
            namespace, resolved = resolve_using_database(
                relative_path=node.path,
                database=self.database,
                location=node.location,
                end_location=node.end_location,
            )

            return replace(node, namespace=namespace, path=resolved)

        return node


def resolve_using_database(
    relative_path: str,
    database: CompilationDatabase,
    location: SourceLocation,
    end_location: SourceLocation,
) -> Tuple[str, str]:
    """Resolve relative resource location."""
    current_file = database.current
    path = database[current_file].resource_location

    if not path:
        exc = InvalidSyntax(
            f"Can't resolve relative resource location {relative_path!r}."
        )
        raise set_location(exc, location, end_location)

    return resolve_relative_location(relative_path, path)


def resolve_relative_location(
    relative_path: str,
    root: str,
    *,
    include_root_file: bool = False,
) -> Tuple[str, str]:
    namespace, _, current_path = root.partition(":")

    resolved = PurePosixPath(current_path)
    if not include_root_file:
        resolved = resolved.parent
    for name in relative_path.split("/"):
        if name == "..":
            resolved = resolved.parent
        elif name and name != ".":
            resolved = resolved / name

    return namespace, str(resolved)



"""Plugin that handles implicit execute commands."""

__all__ = [
    "ImplicitExecuteParser",
]


from dataclasses import dataclass, replace
from typing import Any, Set

from beet import Context
from tokenstream import TokenStream, set_location

from mecha import AstChildren, AstCommand, Mecha, Parser


def beet_default(ctx: Context):
    ctx.require("mecha.contrib.nesting")

    mc = ctx.inject(Mecha)

    if execute := mc.spec.tree.get("execute"):
        commands = {literal for literal, _ in mc.spec.tree.get_all_literals()}
        commands -= {"execute", "run"}

        shorthands = {literal for literal, _ in execute.get_all_literals()}
        shorthands -= {"execute", "run"}

        commands, shorthands = commands - shorthands, shorthands - commands

        mc.spec.add_commands(
            {
                "type": "root",
                "children": {
                    "execute": {
                        "type": "literal",
                        "children": {
                            literal: {"type": "literal", "redirect": [literal]}
                            for literal in commands
                        },
                    },
                    **{
                        literal: {"type": "literal", "redirect": ["execute", literal]}
                        for literal in shorthands
                    },
                },
            }
        )

        mc.spec.parsers["command"] = ImplicitExecuteParser(
            commands={
                name
                for name in mc.spec.prototypes
                if (s := name.split(":", 2)[:2])
                and s[0] == "execute"
                and s[1] in commands
            },
            shorthands={
                name
                for name in mc.spec.prototypes
                if name.partition(":")[0] in shorthands
            },
            parser=mc.spec.parsers["command"],
        )


@dataclass
class ImplicitExecuteParser:
    """Parser that replaces implicit execute shorthands with the full command."""

    commands: Set[str]
    shorthands: Set[str]
    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstCommand):
            if node.identifier in self.commands:
                subcommand = replace(node, identifier=node.identifier[8:])
                run = AstCommand(
                    identifier="execute:run:subcommand",
                    arguments=AstChildren([subcommand]),
                )
                return set_location(run, node)

            elif node.identifier in self.shorthands:
                subcommand = replace(node, identifier=f"execute:{node.identifier}")
                execute = AstCommand(
                    identifier="execute:subcommand",
                    arguments=AstChildren([subcommand]),
                )
                return set_location(execute, node)

        return node



"""Plugin that gathers statistics."""

__all__ = [
    "Analyzer",
    "Summary",
    "Statistics",
]


import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional, Tuple, Union

from beet import Context
from beet.core.utils import dump_json
from pydantic import BaseModel, Field

from mecha import (
    AstCommand,
    AstCommandSentinel,
    AstMacroLine,
    AstMacroLineText,
    AstMacroLineVariable,
    AstObjective,
    AstObjectiveCriteria,
    AstPlayerName,
    AstResourceLocation,
    AstRoot,
    AstSelector,
    AstWord,
    CommandSpec,
    Mecha,
    Reducer,
    rule,
)

logger = logging.getLogger("stats")


class Statistics(BaseModel):
    """Class holding all the stats gathered from the analysis."""

    function_count: int = 0
    command_count: dict[str, dict[str, int]] = Field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )
    command_behind_execute_count: dict[str, int] = Field(
        default_factory=lambda: defaultdict(int)
    )
    macro_line_count: dict[str, int] = Field(default_factory=lambda: defaultdict(int))
    execute_count: int = 0
    execute_clause_count: dict[str, int] = Field(
        default_factory=lambda: defaultdict(int)
    )
    selector_count: dict[str, int] = Field(default_factory=lambda: defaultdict(int))
    selector_entity_type_count: dict[str, int] = Field(
        default_factory=lambda: defaultdict(int)
    )
    selector_argument_count: dict[str, dict[str, int]] = Field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )
    scoreboard_references: dict[str, int] = Field(
        default_factory=lambda: defaultdict(int)
    )
    scoreboard_fake_player_references: dict[str, dict[str, int]] = Field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )
    scoreboard_objectives: Dict[str, str] = {}


class StatisticsOptions(BaseModel):
    output: Optional[str] = None


def beet_default(ctx: Context):
    ctx.inject(Analyzer)


class Analyzer(Reducer):
    """Service that collects command statistics."""

    stats: Statistics

    def __init__(self, ctx: Union[Context, Mecha, None] = None):
        super().__init__()

        if isinstance(ctx, Context):
            ctx.require(self.finalize)
            mc = ctx.inject(Mecha)
        else:
            mc = ctx

        if mc:
            mc.check.extend(self)

        self.stats = Statistics()

    def finalize(self, ctx: Context):
        mc = ctx.inject(Mecha)
        opts = ctx.validate("statistics", StatisticsOptions)
        yield
        logger.info(str(Summary(mc.spec, self.stats)))
        if opts.output:
            output = ctx.directory / opts.output
            output.write_text(dump_json(self.stats.dict()))

    @rule(AstRoot)
    def root(self, node: AstRoot):
        self.stats.function_count += 1

        for command in node.commands:
            if isinstance(command, AstCommandSentinel):
                continue

            behind_execute = False

            while command.identifier.startswith("execute"):
                behind_execute = True
                if command.identifier == "execute:subcommand":
                    self.stats.execute_count += 1
                else:
                    self.stats.execute_clause_count[command.identifier] += 1
                if isinstance(subcommand := command.arguments[-1], AstCommand):
                    command = subcommand
                else:
                    break

            name = command.identifier.partition(":")[0]
            self.stats.command_count[name][command.identifier] += 1

            if behind_execute:
                self.stats.command_behind_execute_count[name] += 1

    @rule(AstSelector)
    def selector(self, node: AstSelector):
        self.stats.selector_count[node.variable] += 1

        for argument in node.arguments:
            if (
                node.variable == "e"
                and argument.key.value == "type"
                and not argument.inverted
                and isinstance(entity_type := argument.value, AstResourceLocation)
            ):
                key = entity_type.get_canonical_value()
                self.stats.selector_entity_type_count[key] += 1
            self.stats.selector_argument_count[node.variable][argument.key.value] += 1

    @rule(AstObjective)
    def objective(self, node: AstObjective):
        self.stats.scoreboard_references[node.value] += 1

    @rule(AstCommand)
    def command(self, node: AstCommand):
        if (
            node.identifier.startswith("scoreboard:objectives:add:objective:criteria")
            and isinstance(objective := node.arguments[0], AstWord)
            and isinstance(criteria := node.arguments[1], AstObjectiveCriteria)
        ):
            self.stats.scoreboard_references[objective.value] += 1
            self.stats.scoreboard_objectives[objective.value] = criteria.value

        ref = None

        for argument in node.arguments:
            if isinstance(argument, AstPlayerName):
                ref = argument.value
            elif ref and isinstance(argument, AstObjective):
                objective = argument.value
                self.stats.scoreboard_fake_player_references[objective][ref] += 1
                ref = None

    @rule(AstMacroLine)
    def macro_line(self, node: AstMacroLine):
        parts = ["$"]
        for arg in node.arguments:
            if isinstance(arg, AstMacroLineText):
                parts.append(arg.value)
            elif isinstance(arg, AstMacroLineVariable):
                parts.append("$(")
                parts.append(arg.value)
                parts.append(")")
        self.stats.macro_line_count["".join(parts)] += 1


@dataclass
class Summary:
    """Formatted statistics summary."""

    spec: CommandSpec
    stats: Statistics

    indent: str = " " * 7

    def __str__(self) -> str:
        return "\n".join(self.format())

    def format(self) -> Iterator[str]:
        plural = "s" * (self.stats.function_count > 1)
        yield f"Analyzed {self.stats.function_count} function{plural}"
        results = {
            **self.format_commands(),
            **self.format_selectors(),
            **self.format_scoreboard(),
        }
        yield from self.format_table(results)

    def format_commands(self) -> Dict[Tuple[str, ...], List[Tuple[str, ...]]]:
        command_stats = {
            (
                f"/{prefix}",
                sum(stats.values()),
                self.stats.command_behind_execute_count[prefix],
            ): [
                (
                    f" {' ' * len(prefix)}{self.spec.prototypes[identifier].usage()[len(prefix) :]}",
                    count,
                    0,
                )
                for identifier, count in stats.items()
            ]
            for prefix, stats in self.stats.command_count.items()
        }

        for macro_line, count in self.stats.macro_line_count.items():
            command_stats[macro_line, count, 0] = []

        if self.stats.execute_count:
            command_stats["/execute", self.stats.execute_count, 0] = [
                (
                    f"         {self.spec.prototypes[clause].usage()}",
                    count,
                    0,
                )
                for clause, count in sorted(
                    self.stats.execute_clause_count.items(), key=lambda p: -p[1]
                )
            ]

        total_commands = sum(total for _, total, _ in command_stats)
        total_commands_behind_execute = sum(total for _, _, total in command_stats)
        total_commands -= total_commands_behind_execute

        return {
            (
                (
                    f"Total commands ({total_commands_behind_execute} behind execute)"
                    if total_commands_behind_execute
                    else "Total commands"
                ),
                str(total_commands),
            ): [
                (
                    (
                        f"{label} ({behind_execute} behind execute)"
                        if behind_execute
                        else label
                    ),
                    str(count),
                )
                for command, stats in sorted(
                    command_stats.items(), key=lambda p: -p[0][1]
                )
                for label, count, behind_execute in [command]
                + (list(sorted(stats, key=lambda p: -p[1])) if len(stats) > 1 else [])
            ],
        }

    def format_selectors(self) -> Dict[Tuple[str, ...], List[Tuple[str, ...]]]:
        selector_types = [
            (f"   [type={entity_type}]", entity_type_count)
            for entity_type, entity_type_count in sorted(
                self.stats.selector_entity_type_count.items(), key=lambda p: -p[1]
            )
        ]

        total_selector_types = sum(self.stats.selector_entity_type_count.values())

        if total_selector_types:
            selector_types.insert(0, ("@e with type", total_selector_types))

        if no_type_count := (self.stats.selector_count["e"] - total_selector_types):
            selector_types.insert(
                0, ("@e with missing or inverted type", no_type_count)
            )

        selector_args = {
            selector: [(f"   [{arg}]", arg_count) for arg, arg_count in args.items()]
            for selector, args in self.stats.selector_argument_count.items()
        }

        return {
            ("Total selectors", str(sum(self.stats.selector_count.values()))): [
                row
                for selector, count in sorted(
                    self.stats.selector_count.items(), key=lambda p: -p[1]
                )
                if count
                for row in [(f"@{selector}", str(count))]
                + [
                    (arg, str(arg_count))
                    for arg, arg_count in sorted(
                        selector_args.get(selector, []), key=lambda p: -p[1]
                    )
                ]
            ]
            + [
                (entity_type, str(entity_type_count))
                for entity_type, entity_type_count in selector_types
            ],
        }

    def format_scoreboard(self) -> Dict[Tuple[str, ...], List[Tuple[str, ...]]]:
        objectives = {
            (
                (
                    f"{objective} ({criteria})"
                    if (criteria := self.stats.scoreboard_objectives.get(objective))
                    else objective
                ),
                count,
            ): [
                (f"{' ' * len(objective)} {fake_player}", fake_player_count)
                for fake_player, fake_player_count in self.stats.scoreboard_fake_player_references[
                    objective
                ].items()
            ]
            for objective, count in self.stats.scoreboard_references.items()
        }

        return {
            ("Scoreboard objectives", str(len(self.stats.scoreboard_references))): [
                (label, str(count))
                for objective, fake_players in sorted(
                    objectives.items(), key=lambda p: -p[0][1]
                )
                for label, count in [objective]
                + list(sorted(fake_players, key=lambda p: -p[1]))
            ]
        }

    def format_table(
        self,
        sections: Dict[Tuple[str, ...], List[Tuple[str, ...]]],
    ) -> Iterator[str]:
        cols = self.compute_column_layout(sections)

        sep = "---".join("-" * col for col in cols)

        for header, rows in sections.items():
            yield sep
            yield self.format_row(header, cols, header=True)
            yield sep

            for row in rows:
                yield self.format_row(row, cols)

    def format_row(
        self,
        row: Tuple[str, ...],
        cols: List[int],
        header: bool = False,
    ) -> str:
        row = ((not header) * self.indent + row[0],) + row[1:]
        return " | ".join(
            (
                (h.rjust(col) if h.isnumeric() else h.ljust(col))
                if len(h) < col
                else h[: col - 3] + "..."
            )
            for h, col in zip(row, cols)
        )

    def compute_column_layout(
        self,
        sections: Dict[Tuple[str, ...], List[Tuple[str, ...]]],
    ) -> List[int]:
        column_count = max(map(len, sections))
        cols: List[int] = []

        for i in range(column_count):
            header_size = max(
                len(header[i]) if len(header) > i else 0 for header in sections
            )
            row_size = max(
                max(len(row[i]) if len(row) > i else 0 for row in rows) if rows else 0
                for rows in sections.values()
            )
            if i == 0:
                row_size += len(self.indent)
            cols.append(min(max(header_size, row_size) + 5, 70))

        return cols



__all__ = [
    "beet_default",
]


from beet import Context

from .api import Mecha


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)

    mc.compile(together=ctx.packs, report=mc.diagnostics)



__all__ = [
    "McdocNode",
    "McdocChildren",
    "McdocDocComment",
    "McdocPrelim",
    "McdocBoolean",
    "McdocInteger",
    "McdocFloat",
    "McdocTypedNumber",
    "McdocIntegerRange",
    "McdocFloatRange",
    "McdocString",
    "McdocResourceLocation",
    "McdocIdentifier",
    "McdocPath",
    "McdocPathSegment",
    "McdocSuper",
    "McdocType",
    "McDocUnattributedType",
    "McdocTypeModifier",
    "McdocAnyType",
    "McdocBooleanType",
    "McdocStringType",
    "McdocNumericType",
    "McdocPrimitiveArrayType",
    "McdocListType",
    "McdocTupleType",
    "McdocEnum",
    "McdocEnumField",
    "McdocEnumValue",
    "McdocStruct",
    "McdocStructField",
    "McdocStructKey",
    "McdocDynamicKey",
    "McdocSpread",
    "McdocReferenceType",
    "McdocDispatcherType",
    "McdocUnionType",
    "McdocIndexBody",
    "McdocIndex",
    "McdocStaticIndex",
    "McdocFallbackIndexKey",
    "McdocNoneIndexKey",
    "McdocUnknownIndexKey",
    "McdocDynamicIndex",
    "McdocAccessor",
    "McdocKeyIndexKey",
    "McdocParentIndexKey",
    "McdocTypeArgBlock",
    "Mcdoc",
    "McdocItem",
    "McdocTypeAlias",
    "McdocUseStatement",
    "McdocDispatchStatement",
    "McdocAttribute",
    "McdocAttributeValue",
    "McdocAttributeTree",
    "McdocAttributeTreeNamedValue",
    "McdocAttributeTreeName",
    "parse_mcdoc",
]

from contextlib import contextmanager
from dataclasses import dataclass, replace
from typing import (
    Callable,
    Iterator,
    List,
    Literal,
    Optional,
    Set,
    TypeAlias,
    TypeVar,
    Union,
)

from beet.core.utils import required_field
from tokenstream import InvalidSyntax, TokenPattern, TokenStream, set_location

from mecha import AbstractChildren, AbstractNode
from mecha.utils import QuoteHelper

McdocNodeType = TypeVar("McdocNodeType", bound="McdocNode", covariant=True)

PATTERN_INTEGER: str = r"[-+]?[1-9][0-9]*|0"
PATTERN_FLOAT: str = r"[-+]?(?:[0-9]*\.[0-9]+|[0-9]+)(?:[eE][-+]?[0-9]+)?"
PATTERN_TYPED_NUMBER: str = rf"{PATTERN_FLOAT}[bBdDfFlLsS]?"
PATTERN_RANGE_DELIMITER: str = r"<?\.\.<?"
PATTERN_INTEGER_RANGE: str = rf"(?:{PATTERN_INTEGER})(?:{PATTERN_RANGE_DELIMITER})(?:{PATTERN_INTEGER})?|(?:{PATTERN_RANGE_DELIMITER})?(?:{PATTERN_INTEGER})"
PATTERN_FLOAT_RANGE: str = rf"(?:{PATTERN_FLOAT})(?:{PATTERN_RANGE_DELIMITER})(?:{PATTERN_FLOAT})?|(?:{PATTERN_RANGE_DELIMITER})?(?:{PATTERN_FLOAT})"
PATTERN_STRING: str = r'"(?:[^"\\]|\\[bfnrt\\"])*"'
PATTERN_RESOURCE_LOCATION: str = r"[a-z0-9_.-]*:(?!:)[a-z0-9_.-]*(?:/[a-z0-9_.-]*)*"
PATTERN_IDENTIFIER: str = r"[a-zA-Z0-9_]+"


RESERVED_WORDS: Set[str] = {
    "any",
    "boolean",
    "byte",
    "double",
    "enum",
    "false",
    "float",
    "int",
    "long",
    "short",
    "string",
    "struct",
    "super",
    "true",
}

MCDOC_QUOTE_HELPER: QuoteHelper = QuoteHelper(
    escape_sequences={
        r"\"": '"',
        r"\\": "\\",
        r"\b": "\b",
        r"\f": "\f",
        r"\r": "\r",
        r"\n": "\n",
        r"\t": "\t",
    }
)


class McdocNode(AbstractNode):
    """Base class for mcdoc nodes."""

    def __iter__(self) -> Iterator["McdocNode"]:
        return super().__iter__()  # pyright: ignore[reportReturnType]

    def walk(self) -> Iterator["McdocNode"]:
        return super().walk()  # pyright: ignore[reportReturnType]


class McdocChildren(AbstractChildren[McdocNodeType]):
    """Specialized tuple subclass for holding multiple child mcdoc nodes."""


@dataclass(frozen=True, slots=True)
class McdocDocComment(McdocNode):
    """Mcdoc doc comment node."""

    text: str = required_field()


@dataclass(frozen=True, slots=True)
class McdocPrelim(McdocNode):
    """Mcdoc prelim node."""

    doc_comments: McdocChildren[McdocDocComment] = required_field()
    attributes: McdocChildren["McdocAttribute"] = required_field()


@dataclass(frozen=True, slots=True)
class McdocBoolean(McdocNode):
    """Mcdoc boolean node."""

    value: bool = required_field()


@dataclass(frozen=True, slots=True)
class McdocInteger(McdocNode):
    """Mcdoc integer node."""

    value: int = required_field()


@dataclass(frozen=True, slots=True)
class McdocFloat(McdocNode):
    """Mcdoc float node."""

    value: float = required_field()


@dataclass(frozen=True, slots=True)
class McdocTypedNumber(McdocNode):
    """Mcdoc typed number node."""

    value: Union[int, float] = required_field()
    type: Literal["byte", "short", "int", "long", "float", "double"] = required_field()


@dataclass(frozen=True, slots=True)
class McdocIntegerRange(McdocNode):
    """Mcdoc integer range node."""

    left: Optional[int] = required_field()
    delimiter: Literal["..", "..<", "<..", "<..<"] = required_field()
    right: Optional[int] = required_field()


@dataclass(frozen=True, slots=True)
class McdocFloatRange(McdocNode):
    """Mcdoc float range node."""

    left: Optional[float] = required_field()
    delimiter: Literal["..", "..<", "<..", "<..<"] = required_field()
    right: Optional[float] = required_field()


@dataclass(frozen=True, slots=True)
class McdocString(McdocNode):
    """Mcdoc string node."""

    value: str = required_field()


@dataclass(frozen=True, slots=True)
class McdocResourceLocation(McdocNode):
    """Mcdoc resource location node."""

    value: str = required_field()


@dataclass(frozen=True, slots=True)
class McdocIdentifier(McdocNode):
    """Mcdoc identifier node."""

    name: str = required_field()


@dataclass(frozen=True, slots=True)
class McdocPath(McdocNode):
    """Mcdoc path node."""

    absolute: bool = required_field()
    segments: McdocChildren["McdocPathSegment"] = required_field()


McdocPathSegment: TypeAlias = Union[McdocIdentifier, "McdocSuper"]


@dataclass(frozen=True, slots=True)
class McdocSuper(McdocNode):
    """Mcdoc super node."""


@dataclass(frozen=True, slots=True)
class McdocType(McdocNode):
    """Mcdoc type node."""

    attributes: McdocChildren["McdocAttribute"] = required_field()
    unattributed_type: "McDocUnattributedType" = required_field()
    modifiers: McdocChildren["McdocTypeModifier"] = required_field()


McDocUnattributedType: TypeAlias = Union[
    "McdocAnyType",
    "McdocBooleanType",
    "McdocStringType",
    McdocBoolean,
    McdocString,
    McdocTypedNumber,
    "McdocNumericType",
    "McdocPrimitiveArrayType",
    "McdocListType",
    "McdocTupleType",
    "McdocEnum",
    "McdocStruct",
    "McdocReferenceType",
    "McdocDispatcherType",
    "McdocUnionType",
]

McdocTypeModifier: TypeAlias = Union["McdocIndexBody", "McdocTypeArgBlock"]


@dataclass(frozen=True, slots=True)
class McdocAnyType(McdocNode):
    """Mcdoc any type node."""


@dataclass(frozen=True, slots=True)
class McdocBooleanType(McdocNode):
    """Mcdoc boolean type node."""


@dataclass(frozen=True, slots=True)
class McdocStringType(McdocNode):
    """Mcdoc string type node."""

    length: Optional[McdocIntegerRange] = None


@dataclass(frozen=True, slots=True)
class McdocNumericType(McdocNode):
    """Mcdoc numeric type node."""

    name: Literal["byte", "short", "int", "long", "float", "double"] = required_field()
    constraint: Optional[McdocIntegerRange] = None


@dataclass(frozen=True, slots=True)
class McdocPrimitiveArrayType(McdocNode):
    """Mcdoc primitive array type node."""

    name: Literal["byte", "int", "long"] = required_field()
    constraint: Optional[McdocIntegerRange] = None
    length: Optional[McdocIntegerRange] = None


@dataclass(frozen=True, slots=True)
class McdocListType(McdocNode):
    """Mcdoc list type node."""

    element: McdocType = required_field()
    length: Optional[McdocIntegerRange] = None


@dataclass(frozen=True, slots=True)
class McdocTupleType(McdocNode):
    """Mcdoc tuple type node."""

    elements: McdocChildren[McdocType] = required_field()


@dataclass(frozen=True, slots=True)
class McdocEnum(McdocNode):
    """Mcdoc enum node."""

    prelim: McdocPrelim = required_field()
    type: Literal[
        "byte",
        "short",
        "int",
        "long",
        "string",
        "float",
        "double",
    ] = required_field()
    identifier: Optional[McdocIdentifier] = required_field()
    fields: McdocChildren["McdocEnumField"] = required_field()


@dataclass(frozen=True, slots=True)
class McdocEnumField(McdocNode):
    """Mcdoc enum field node."""

    prelim: McdocPrelim = required_field()
    identifier: McdocIdentifier = required_field()
    value: "McdocEnumValue" = required_field()


McdocEnumValue: TypeAlias = Union[McdocTypedNumber, McdocString]


@dataclass(frozen=True, slots=True)
class McdocStruct(McdocNode):
    """Mcdoc struct node."""

    prelim: McdocPrelim = required_field()
    identifier: Optional[McdocIdentifier] = required_field()
    fields: McdocChildren[Union["McdocStructField", "McdocSpread"]] = required_field()


@dataclass(frozen=True, slots=True)
class McdocStructField(McdocNode):
    """Mcdoc struct field node."""

    prelim: McdocPrelim = required_field()
    key: "McdocStructKey" = required_field()
    optional: bool = required_field()
    type: McdocType = required_field()


McdocStructKey: TypeAlias = Union[McdocString, McdocIdentifier, "McdocDynamicKey"]


@dataclass(frozen=True, slots=True)
class McdocDynamicKey(McdocNode):
    """Mcdoc dynamic key node."""

    type: McdocType = required_field()


@dataclass(frozen=True, slots=True)
class McdocSpread(McdocNode):
    """Mcdoc spread node."""

    attributes: McdocChildren["McdocAttribute"] = required_field()
    type: McdocType = required_field()


@dataclass(frozen=True, slots=True)
class McdocReferenceType(McdocNode):
    """Mcdoc reference type node."""

    path: McdocPath = required_field()


@dataclass(frozen=True, slots=True)
class McdocDispatcherType(McdocNode):
    """Mcdoc dispatcher type node."""

    resource_location: McdocResourceLocation = required_field()
    indices: McdocChildren["McdocIndex"] = required_field()


@dataclass(frozen=True, slots=True)
class McdocUnionType(McdocNode):
    """Mcdoc union type node."""

    types: McdocChildren[McdocType] = required_field()


@dataclass(frozen=True, slots=True)
class McdocIndexBody(McdocNode):
    """Mcdoc index body node."""

    indices: McdocChildren["McdocIndex"] = required_field()


McdocIndex: TypeAlias = Union[
    "McdocStaticIndex",
    "McdocDynamicIndex",
]

McdocStaticIndex: TypeAlias = Union[
    "McdocFallbackIndexKey",
    "McdocNoneIndexKey",
    "McdocUnknownIndexKey",
    McdocIdentifier,
    McdocString,
    McdocResourceLocation,
]


@dataclass(frozen=True, slots=True)
class McdocFallbackIndexKey(McdocNode):
    """Mcdoc %fallback index key node."""


@dataclass(frozen=True, slots=True)
class McdocNoneIndexKey(McdocNode):
    """Mcdoc %none index key node."""


@dataclass(frozen=True, slots=True)
class McdocUnknownIndexKey(McdocNode):
    """Mcdoc %unknown index key node."""


@dataclass(frozen=True, slots=True)
class McdocDynamicIndex(McdocNode):
    """Mcdoc dynamic index node."""

    accessors: McdocChildren["McdocAccessor"] = required_field()


McdocAccessor: TypeAlias = Union[
    "McdocKeyIndexKey",
    "McdocParentIndexKey",
    McdocIdentifier,
    McdocString,
]


@dataclass(frozen=True, slots=True)
class McdocKeyIndexKey(McdocNode):
    """Mcdoc %key index key node."""


@dataclass(frozen=True, slots=True)
class McdocParentIndexKey(McdocNode):
    """Mcdoc %parent index key node."""


@dataclass(frozen=True, slots=True)
class McdocTypeArgBlock(McdocNode):
    """Mcdoc type arg block node."""

    arguments: McdocChildren[McdocType] = required_field()


@dataclass(frozen=True, slots=True)
class Mcdoc(McdocNode):
    """Mcdoc node."""

    items: McdocChildren["McdocItem"] = required_field()


McdocItem: TypeAlias = Union[
    McdocStruct,
    McdocEnum,
    "McdocTypeAlias",
    "McdocUseStatement",
    "McdocDispatchStatement",
]


@dataclass(frozen=True, slots=True)
class McdocTypeAlias(McdocNode):
    """Mcdoc type alias node."""

    prelim: McdocPrelim = required_field()
    identifier: McdocIdentifier = required_field()
    parameters: McdocChildren[McdocIdentifier] = required_field()
    type: McdocType = required_field()


@dataclass(frozen=True, slots=True)
class McdocUseStatement(McdocNode):
    """Mcdoc use statement node."""

    path: McdocPath = required_field()
    alias: Optional[McdocIdentifier] = None


@dataclass(frozen=True, slots=True)
class McdocDispatchStatement(McdocNode):
    """Mcdoc dispatch statement node."""

    attributes: McdocChildren["McdocAttribute"] = required_field()
    resource_location: McdocResourceLocation = required_field()
    indices: McdocChildren["McdocStaticIndex"] = required_field()
    parameters: McdocChildren[McdocIdentifier] = required_field()
    type: McdocType = required_field()


@dataclass(frozen=True, slots=True)
class McdocAttribute(McdocNode):
    """Mcdoc attribute node."""

    identifier: McdocIdentifier = required_field()
    value: Optional["McdocAttributeValue"] = None


McdocAttributeValue: TypeAlias = Union[McdocType, "McdocAttributeTree"]


@dataclass(frozen=True, slots=True)
class McdocAttributeTree(McdocNode):
    """Mcdoc attribute tree node."""

    flavor: Literal["brace", "bracket", "curly"] = required_field()
    positional_values: McdocChildren[McdocAttributeValue] = required_field()
    named_values: McdocChildren["McdocAttributeTreeNamedValue"] = required_field()


@dataclass(frozen=True, slots=True)
class McdocAttributeTreeNamedValue(McdocNode):
    """Mcdoc attribute tree named value node."""

    name: "McdocAttributeTreeName" = required_field()
    value: McdocAttributeValue = required_field()


McdocAttributeTreeName: TypeAlias = Union[McdocIdentifier, McdocString]


@contextmanager
def ignore_comments(stream: TokenStream):
    with stream.syntax(comment=r"//(?!/)[^\n]*\n?"), stream.ignore("comment"):
        yield


def parse_prelim(stream: TokenStream) -> McdocPrelim:
    with stream.syntax(doc_comment=r"///[^\n]*\n?"):
        with stream.intercept("comment"):
            while stream.get("comment"):
                pass

            start = stream.peek()
            doc_comments: List[McdocDocComment] = []

            for token in stream.collect("doc_comment"):
                node = McdocDocComment(text=token.value)
                doc_comments.append(set_location(node, token))

        attributes = parse_attributes(stream)

        node = McdocPrelim(
            doc_comments=McdocChildren(doc_comments),
            attributes=attributes,
        )
        return set_location(
            node,
            start,
            stream.current if doc_comments or attributes else start,
        )


def parse_integer(stream: TokenStream) -> McdocInteger:
    with stream.syntax(integer=PATTERN_INTEGER):
        token = stream.expect("integer")
        node = McdocInteger(value=int(token.value))
        return set_location(node, token)


def parse_float(stream: TokenStream) -> McdocFloat:
    with stream.syntax(float=PATTERN_FLOAT):
        token = stream.expect("float")
        node = McdocFloat(value=float(token.value))
        return set_location(node, token)


def parse_typed_number(stream: TokenStream) -> McdocTypedNumber:
    with stream.syntax(typed_number=PATTERN_TYPED_NUMBER):
        token = stream.expect("typed_number")
        normalized = token.value.lower()

        if normalized.endswith(("b", "s", "l")):
            if "." in normalized:
                value = int(float(normalized[:-1]))
            else:
                value = int(normalized[:-1])
            suffix = normalized[-1]
            type = "byte" if suffix == "b" else "short" if suffix == "s" else "long"
        elif normalized.endswith(("f", "d")):
            value = float(normalized[:-1])
            suffix = normalized[-1]
            type = "float" if suffix == "f" else "double"
        elif "." in normalized:
            value = float(normalized)
            type = "double"
        else:
            value = int(normalized)
            type = "int"

        node = McdocTypedNumber(value=value, type=type)
        return set_location(node, token)


def parse_integer_range(stream: TokenStream) -> McdocIntegerRange:
    with stream.syntax(integer_range=PATTERN_INTEGER_RANGE):
        token = stream.expect("integer_range")
        before, delimiter, after = token.value.partition("..")

        if delimiter:
            left = int(before.removesuffix("<")) if before else None
            right = int(after.removeprefix("<")) if after else None
        else:
            left = int(before)
            delimiter = ".."
            right = left

        node = McdocIntegerRange(left=left, delimiter=delimiter, right=right)  # pyright: ignore[reportArgumentType]
        return set_location(node, token)


def parse_float_range(stream: TokenStream) -> McdocFloatRange:
    with stream.syntax(float_range=PATTERN_FLOAT_RANGE):
        token = stream.expect("float_range")
        before, delimiter, after = token.value.partition("..")

        if delimiter:
            left = float(before.removesuffix("<")) if before else None
            right = float(after.removeprefix("<")) if after else None
        else:
            left = float(before)
            delimiter = ".."
            right = left

        node = McdocFloatRange(left=left, delimiter=delimiter, right=right)  # pyright: ignore[reportArgumentType]
        return set_location(node, token)


def parse_string(stream: TokenStream) -> McdocString:
    with stream.syntax(string=PATTERN_STRING):
        token = stream.expect("string")
        value = MCDOC_QUOTE_HELPER.unquote_string(token)
        node = McdocString(value=value)
        return set_location(node, token)


def parse_resource_location(stream: TokenStream) -> McdocResourceLocation:
    with stream.syntax(resource_location=PATTERN_RESOURCE_LOCATION):
        token = stream.expect("resource_location")
        node = McdocResourceLocation(value=token.value)
        return set_location(node, token)


def parse_identifier(stream: TokenStream) -> McdocIdentifier:
    with stream.syntax(identifier=PATTERN_IDENTIFIER):
        token = stream.expect("identifier")
        name = token.value

        if name in RESERVED_WORDS:
            exc = InvalidSyntax(f'Can\'t use reserved word "{name}" as an identifier.')
            raise set_location(exc, token)

        node = McdocIdentifier(name=name)
        return set_location(node, token)


def parse_path(stream: TokenStream) -> McdocPath:
    with stream.syntax(delimiter=r"::", super=r"super\b"):
        start = stream.get("delimiter")

        absolute = bool(start)
        segments: List[McdocPathSegment] = []

        while True:
            if token := stream.get("super"):
                node = set_location(McdocSuper(), token)
            else:
                node = parse_identifier(stream)
            segments.append(node)
            if not stream.get("delimiter"):
                break

        node = McdocPath(absolute=absolute, segments=McdocChildren(segments))
        return set_location(node, start or segments[0], segments[-1])


def parse_type(stream: TokenStream) -> McdocType:
    with stream.syntax(bracket=r"\[|\]", angle=r"<|>"):
        attributes = parse_attributes(stream)

        unattributed_type: McDocUnattributedType = parse_unattributed_type(stream)
        modifiers: List[McdocTypeModifier] = []

        for open_bracket, open_angle in stream.collect(
            ("bracket", "["),
            ("angle", "<"),
        ):
            if open_bracket:
                indices: List[McdocIndex] = []
                close_bracket = expect_sequence(
                    stream,
                    indices,
                    lambda s: choose_alternative(
                        s, parse_dynamic_index, parse_static_index
                    ),
                    until=("bracket", "]"),
                )
                if not indices:
                    exc = InvalidSyntax("Expected at least one index.")
                    raise set_location(exc, open_bracket, close_bracket)
                node = McdocIndexBody(indices=McdocChildren(indices))
                modifiers.append(set_location(node, open_bracket, close_bracket))

            elif open_angle:
                arguments: List[McdocType] = []
                close_angle = expect_sequence(
                    stream,
                    arguments,
                    parse_type,
                    until=("angle", ">"),
                )
                node = McdocTypeArgBlock(arguments=McdocChildren(arguments))
                modifiers.append(set_location(node, open_angle, close_angle))

    node = McdocType(
        attributes=attributes,
        unattributed_type=unattributed_type,
        modifiers=McdocChildren(modifiers),
    )
    return set_location(
        node,
        attributes[0] if attributes else unattributed_type,
        modifiers[-1] if modifiers else unattributed_type,
    )


def parse_unattributed_type(stream: TokenStream) -> McDocUnattributedType:
    with stream.syntax(
        any=r"any\b",
        boolean=r"boolean\b",
        string=r"string\b",
        constraint=r"@",
        true=r"true\b",
        false=r"false\b",
        byte=r"byte\b",
        short=r"short\b",
        int=r"int\b",
        long=r"long\b",
        float=r"float\b",
        double=r"double\b",
        array=r"\[\]",
        bracket=r"\[|\]",
        comma=r",",
    ):
        token = stream.get(
            "any",
            "boolean",
            "string",
            "true",
            "false",
            "byte",
            "short",
            "int",
            "long",
            "float",
            "double",
            ("bracket", "["),
        )

        if not token:
            return choose_alternative(
                stream,
                parse_string,
                parse_typed_number,
                parse_enum,
                parse_struct,
                parse_dispatcher_type,
                parse_reference_type,
                parse_union_type,
            )

        if token.match("any"):
            return set_location(McdocAnyType(), token)
        elif token.match("boolean"):
            return set_location(McdocBooleanType(), token)
        elif token.match("string"):
            length = parse_integer_range(stream) if stream.get("constraint") else None
            node = McdocStringType(length=length)
            return set_location(node, token, node.length)
        elif token.match("true"):
            return set_location(McdocBoolean(value=True), token)
        elif token.match("false"):
            return set_location(McdocBoolean(value=False), token)

        if token.type in ("byte", "short", "int", "long", "float", "double"):
            node = set_location(McdocNumericType(name=token.type), token)

            if stream.get("constraint"):
                node = replace(node, constraint=parse_integer_range(stream))
                node = set_location(node, node, node.constraint)

            if node.name in ("byte", "int", "long"):
                if array := stream.get("array"):
                    length = (
                        parse_integer_range(stream)
                        if stream.get("constraint")
                        else None
                    )
                    node = McdocPrimitiveArrayType(
                        name=node.name,
                        constraint=node.constraint,
                        length=length,
                    )
                    node = set_location(node, node.name, length or array)

            return node

        element = parse_type(stream)

        if close_bracket := stream.get(("bracket", "]")):
            length = None
            if stream.get("constraint"):
                length = parse_integer_range(stream)
            if length == McdocIntegerRange(left=1, delimiter="..", right=1):
                node = McdocTupleType(elements=McdocChildren([element]))
            else:
                node = McdocListType(element=element, length=length)
            return set_location(node, token, length or close_bracket)

        stream.expect("comma")

        elements = [element]
        close_bracket = expect_sequence(
            stream,
            elements,
            parse_type,
            until=("bracket", "]"),
        )
        node = McdocTupleType(elements=McdocChildren(elements))
        return set_location(node, token, close_bracket)


def parse_enum(stream: TokenStream) -> McdocEnum:
    with stream.syntax(
        enum=r"enum\b",
        brace=r"\(|\)",
        byte=r"byte\b",
        short=r"short\b",
        int=r"int\b",
        long=r"long\b",
        string=r"string\b",
        float=r"float\b",
        double=r"double\b",
        curly=r"\{|\}",
    ):
        prelim = parse_prelim(stream)

        stream.expect("enum")
        stream.expect(("brace", "("))
        type = stream.expect_any(
            "byte",
            "short",
            "int",
            "long",
            "string",
            "float",
            "double",
        )
        stream.expect(("brace", ")"))

        identifier = None
        with stream.alternative():
            identifier = parse_identifier(stream)

        fields: List[McdocEnumField] = []
        stream.expect(("curly", "{"))
        close_curly = expect_sequence(
            stream,
            fields,
            parse_enum_field,
            until=("curly", "}"),
        )

        node = McdocEnum(
            prelim=prelim,
            type=type.value,  # pyright: ignore[reportArgumentType]
            identifier=identifier,
            fields=McdocChildren(fields),
        )
        return set_location(node, prelim, close_curly)


def parse_enum_field(stream: TokenStream) -> McdocEnumField:
    with stream.syntax(equal=r"="):
        prelim = parse_prelim(stream)

        identifier = parse_identifier(stream)
        stream.expect("equal")
        value = choose_alternative(stream, parse_string, parse_typed_number)

        node = McdocEnumField(prelim=prelim, identifier=identifier, value=value)
        return set_location(node, prelim, value)


def parse_struct(stream: TokenStream) -> McdocStruct:
    with stream.syntax(struct=r"struct\b", curly=r"\{|\}"):
        prelim = parse_prelim(stream)

        stream.expect("struct")

        identifier = None
        with stream.alternative():
            identifier = parse_identifier(stream)

        fields: List[Union[McdocStructField, McdocSpread]] = []
        stream.expect(("curly", "{"))
        close_curly = expect_sequence(
            stream,
            fields,
            lambda s: choose_alternative(s, parse_struct_field, parse_spread),
            until=("curly", "}"),
        )

        node = McdocStruct(
            prelim=prelim,
            identifier=identifier,
            fields=McdocChildren(fields),
        )
        return set_location(node, prelim, close_curly)


def parse_struct_field(stream: TokenStream) -> McdocStructField:
    with stream.syntax(optional=r"\?", colon=r":", bracket=r"\[|\]"):
        prelim = parse_prelim(stream)

        if open_bracket := stream.get(("bracket", "[")):
            type = parse_type(stream)
            close_bracket = stream.expect(("bracket", "]"))
            key = McdocDynamicKey(type=type)
            key = set_location(key, open_bracket, close_bracket)
        else:
            key = choose_alternative(stream, parse_string, parse_identifier)

        optional = bool(stream.get("optional"))
        stream.expect("colon")

        type = parse_type(stream)

        node = McdocStructField(prelim=prelim, key=key, optional=optional, type=type)
        return set_location(node, prelim, type)


def parse_spread(stream: TokenStream) -> McdocSpread:
    with stream.syntax(spread=r"\.\.\."):
        attributes = parse_attributes(stream)
        token = stream.expect("spread")
        type = parse_type(stream)
        node = McdocSpread(attributes=attributes, type=type)
        return set_location(node, attributes[0] if attributes else token, type)


def parse_reference_type(stream: TokenStream) -> McdocReferenceType:
    path = parse_path(stream)
    node = McdocReferenceType(path=path)
    return set_location(node, path)


def parse_dispatcher_type(stream: TokenStream) -> McdocDispatcherType:
    with stream.syntax(bracket=r"\[|\]"):
        resource_location = parse_resource_location(stream)

        open_bracket = stream.expect(("bracket", "["))
        indices: List[McdocIndex] = []
        close_bracket = expect_sequence(
            stream,
            indices,
            lambda s: choose_alternative(s, parse_dynamic_index, parse_static_index),
            until=("bracket", "]"),
        )

        if not indices:
            exc = InvalidSyntax("Expected at least one index.")
            raise set_location(exc, open_bracket, close_bracket)

        node = McdocDispatcherType(
            resource_location=resource_location,
            indices=McdocChildren(indices),
        )
        return set_location(node, resource_location, close_bracket)


def parse_union_type(stream: TokenStream) -> McdocUnionType:
    with stream.syntax(brace=r"\(|\)", pipe=r"\|"):
        open_brace = stream.expect(("brace", "("))
        types: List[McdocType] = []
        close_brace = expect_sequence(
            stream,
            types,
            parse_type,
            until=("brace", ")"),
            separator=r"\|",
        )
        node = McdocUnionType(types=McdocChildren(types))
        return set_location(node, open_brace, close_brace)


def parse_static_index(stream: TokenStream) -> McdocStaticIndex:
    with stream.syntax(fallback=r"%fallback\b", none=r"%none\b", unknown=r"%unknown\b"):
        with stream.alternative():
            fallback, none, unknown = stream.expect("fallback", "none", "unknown")
            if fallback:
                return set_location(McdocFallbackIndexKey(), fallback)
            elif none:
                return set_location(McdocNoneIndexKey(), none)
            elif unknown:
                return set_location(McdocUnknownIndexKey(), unknown)

        return choose_alternative(
            stream,
            parse_string,
            parse_resource_location,
            parse_identifier,
        )


def parse_dynamic_index(stream: TokenStream) -> McdocDynamicIndex:
    with stream.syntax(bracket=r"\[|\]", dot=r"\."):
        open_bracket = stream.expect(("bracket", "["))

        accessors: List[McdocAccessor] = [parse_accessor(stream)]
        for _ in stream.collect("dot"):
            accessors.append(parse_accessor(stream))

        close_bracket = stream.expect(("bracket", "]"))

        node = McdocDynamicIndex(accessors=McdocChildren(accessors))
        return set_location(node, open_bracket, close_bracket)


def parse_accessor(stream: TokenStream) -> McdocAccessor:
    with stream.syntax(key=r"%key\b", parent=r"%parent\b"):
        with stream.alternative():
            key, parent = stream.expect("key", "parent")
            if key:
                return set_location(McdocKeyIndexKey(), key)
            elif parent:
                return set_location(McdocParentIndexKey(), parent)

        return choose_alternative(
            stream,
            parse_string,
            parse_identifier,
        )


def parse_mcdoc(stream: TokenStream) -> Mcdoc:
    with ignore_comments(stream):
        items: List[McdocItem] = []

        while stream.peek():
            item = choose_alternative(
                stream,
                parse_struct,
                parse_enum,
                parse_type_alias,
                parse_use_statement,
                parse_dispatch_statement,
            )
            items.append(item)

        node = Mcdoc(items=McdocChildren(items))
        return set_location(
            node,
            items[0] if items else stream.current,
            items[-1] if items else stream.current,
        )


def parse_type_alias(stream: TokenStream) -> McdocTypeAlias:
    with stream.syntax(type=r"type\b", angle=r"<|>", equal=r"="):
        prelim = parse_prelim(stream)

        stream.expect("type")
        identifier = parse_identifier(stream)

        parameters: List[McdocIdentifier] = []
        if stream.get(("angle", "<")):
            expect_sequence(
                stream,
                parameters,
                parse_identifier,
                until=("angle", ">"),
            )

        stream.expect("equal")

        type = parse_type(stream)

        node = McdocTypeAlias(
            prelim=prelim,
            identifier=identifier,
            parameters=McdocChildren(parameters),
            type=type,
        )
        return set_location(node, prelim, type)


def parse_use_statement(stream: TokenStream) -> McdocUseStatement:
    with stream.syntax(use=r"use\b", as_=r"as\b"):
        token = stream.expect("use")
        path = parse_path(stream)

        alias = None
        if stream.get("as_"):
            alias = parse_identifier(stream)

        node = McdocUseStatement(path=path, alias=alias)
        return set_location(node, token, alias or path)


def parse_dispatch_statement(stream: TokenStream) -> McdocDispatchStatement:
    with stream.syntax(
        dispatch=r"dispatch\b",
        bracket=r"\[|\]",
        angle=r"<|>",
        to=r"to\b",
    ):
        attributes = parse_attributes(stream)

        token = stream.expect("dispatch")

        resource_location = parse_resource_location(stream)

        open_bracket = stream.expect(("bracket", "["))
        indices: List[McdocStaticIndex] = []
        close_bracket = expect_sequence(
            stream,
            indices,
            parse_static_index,
            until=("bracket", "]"),
        )

        if not indices:
            exc = InvalidSyntax("Expected at least one index.")
            raise set_location(exc, open_bracket, close_bracket)

        parameters: List[McdocIdentifier] = []
        if stream.get(("angle", "<")):
            expect_sequence(
                stream,
                parameters,
                parse_identifier,
                until=("angle", ">"),
            )

        stream.expect("to")

        type = parse_type(stream)

        node = McdocDispatchStatement(
            attributes=attributes,
            resource_location=resource_location,
            indices=McdocChildren(indices),
            parameters=McdocChildren(parameters),
            type=type,
        )
        return set_location(node, attributes[0] if attributes else token, type)


def parse_attributes(stream: TokenStream) -> McdocChildren[McdocAttribute]:
    with stream.syntax(attribute=r"#\[|\]", equal=r"="):
        attributes: List[McdocAttribute] = []

        for open_attribute in stream.collect(("attribute", "#[")):
            identifier = parse_identifier(stream)

            if close_attribute := stream.get(("attribute", "]")):
                value = None
            else:
                if stream.get("equal"):
                    value = choose_alternative(stream, parse_type, parse_attribute_tree)
                else:
                    value = parse_attribute_tree(stream)
                close_attribute = stream.expect(("attribute", "]"))

            node = McdocAttribute(identifier=identifier, value=value)
            attributes.append(set_location(node, open_attribute, close_attribute))

        return McdocChildren(attributes)


def parse_attribute_tree(stream: TokenStream) -> McdocAttributeTree:
    with stream.syntax(brace=r"\(|\)", bracket=r"\[|\]", curly=r"\{|\}", comma=r","):
        open_tree = stream.expect_any(
            ("brace", "("),
            ("bracket", "["),
            ("curly", "{"),
        )

        close_pattern = (
            ("brace", ")")
            if open_tree.match("brace")
            else ("bracket", "]")
            if open_tree.match("bracket")
            else ("curly", "}")
        )

        positional_values: List[McdocAttributeValue] = []
        named_values: List[McdocAttributeTreeNamedValue] = []

        for _ in stream.peek_until(close_pattern):
            value = choose_alternative(
                stream,
                parse_attribute_tree,
                parse_attribute_tree_named_value,
                parse_type,
            )

            if isinstance(value, McdocAttributeTreeNamedValue):
                named_values.append(value)
            elif not named_values:
                positional_values.append(value)
            else:
                exc = InvalidSyntax("Invalid positional value after named value.")
                raise set_location(exc, value)

            if not stream.get("comma"):
                stream.expect(close_pattern)
                break

        node = McdocAttributeTree(
            flavor=close_pattern[0],
            positional_values=McdocChildren(positional_values),
            named_values=McdocChildren(named_values),
        )
        return set_location(node, open_tree, stream.current)


def parse_attribute_tree_named_value(
    stream: TokenStream,
) -> McdocAttributeTreeNamedValue:
    with stream.syntax(equal=r"="):
        name = choose_alternative(stream, parse_string, parse_identifier)

        if stream.get("equal"):
            value = choose_alternative(stream, parse_type, parse_attribute_tree)
        else:
            value = parse_attribute_tree(stream)

        node = McdocAttributeTreeNamedValue(name=name, value=value)
        return set_location(node, name, value)


def expect_sequence(
    stream: TokenStream,
    nodes: List[McdocNodeType],
    parser: Callable[[TokenStream], McdocNodeType],
    until: TokenPattern,
    separator: str = r",",
):
    with stream.syntax(separator=separator):
        for _ in stream.peek_until(until):
            nodes.append(parser(stream))
            if not stream.get("separator"):
                stream.expect(until)
                break
        return stream.current


def choose_alternative(
    stream: TokenStream,
    *args: Callable[[TokenStream], McdocNodeType],
) -> McdocNodeType:
    for parser, alternative in stream.choose(*args):
        with alternative:
            return parser(stream)
    raise ValueError("Unreachable!")



"""Plugin for handling nested yaml."""

__all__ = [
    "BaseYamlObjectCollector",
    "BaseYamlArrayCollector",
    "JsonObjectCollector",
    "JsonArrayCollector",
    "NbtCompoundCollector",
    "NbtListCollector",
    "NestedYamlParser",
    "NestedYamlContextProvider",
    "collect_json_string",
    "collect_nbt_string",
    "YAML_STRING_PATTERN",
]


from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Any, Callable, Generic, Iterator, List, TypeVar

from beet import Context
from nbtlib import String
from tokenstream import SourceLocation, TokenStream, set_location

from mecha import (
    AstChildren,
    AstJson,
    AstJsonArray,
    AstJsonObject,
    AstJsonObjectEntry,
    AstJsonObjectKey,
    AstJsonValue,
    AstNbt,
    AstNbtCompound,
    AstNbtCompoundEntry,
    AstNbtCompoundKey,
    AstNbtList,
    AstNbtValue,
    Mecha,
    Parser,
    consume_line_continuation,
    delegate,
)
from mecha.utils import JsonQuoteHelper, QuoteHelper

EntryType = TypeVar("EntryType")
ValueType = TypeVar("ValueType")


YAML_STRING_PATTERN = r'"(?:\\.|[^\\\n])*?"' "|" r"'(?:\\.|[^\\\n])*?'"


def beet_default(ctx: Context):
    ctx.require("mecha.contrib.nesting")

    mc = ctx.inject(Mecha)

    mc.spec.parsers["json"] = NestedYamlParser(
        parser=delegate("json"),
        original_parser=mc.spec.parsers["json"],
        object_collector=JsonObjectCollector(),
        array_collector=JsonArrayCollector(),
        string_collector=collect_json_string,
    )

    mc.spec.parsers["nbt"] = NestedYamlParser(
        parser=delegate("nbt"),
        original_parser=mc.spec.parsers["nbt"],
        object_collector=NbtCompoundCollector(),
        array_collector=NbtListCollector(),
        string_collector=collect_nbt_string,
    )

    for argument_name in [
        "command:argument:minecraft:component",
        "command:argument:minecraft:nbt_compound_tag",
        "command:argument:minecraft:nbt_tag",
        "command:argument:mecha:nested_json",
    ]:
        if parser := mc.spec.parsers.get(argument_name):
            mc.spec.parsers[argument_name] = NestedYamlContextProvider(parser)


@dataclass
class BaseYamlObjectCollector(Generic[EntryType, ValueType]):
    """Base class for yaml object collectors."""

    entries: List[EntryType] = field(default_factory=list)

    @contextmanager
    def start(self) -> Iterator[None]:
        """Start a new object."""
        previous_entries = self.entries
        self.entries = []
        try:
            yield
        finally:
            self.entries = previous_entries

    def emit(
        self,
        key: str,
        value: ValueType,
        location: SourceLocation,
        end_location: SourceLocation,
    ) -> None:
        """Emit an object entry."""
        raise NotImplementedError()

    def result(self) -> ValueType:
        """Output the final ast node."""
        raise NotImplementedError()


@dataclass
class BaseYamlArrayCollector(Generic[ValueType]):
    """Base class for yaml array collectors."""

    elements: List[ValueType] = field(default_factory=list)

    @contextmanager
    def start(self) -> Iterator[None]:
        """Start a new array."""
        previous_elements = self.elements
        self.elements = []
        try:
            yield
        finally:
            self.elements = previous_elements

    def emit(self, value: ValueType) -> None:
        """Emit an array element."""
        raise NotImplementedError()

    def result(self) -> ValueType:
        """Output the final ast node."""
        raise NotImplementedError()


class JsonObjectCollector(BaseYamlObjectCollector[AstJsonObjectEntry, AstJson]):
    """Collect yaml as json objects."""

    def emit(
        self,
        key: str,
        value: AstJson,
        location: SourceLocation,
        end_location: SourceLocation,
    ) -> None:
        key_node = AstJsonObjectKey(value=key)
        key_node = set_location(key_node, location, end_location)
        entry_node = AstJsonObjectEntry(key=key_node, value=value)
        self.entries.append(set_location(entry_node, key_node, value))

    def result(self) -> AstJson:
        node = AstJsonObject(entries=AstChildren(self.entries))
        return set_location(node, node.entries[0], node.entries[-1])


class JsonArrayCollector(BaseYamlArrayCollector[AstJson]):
    """Collect yaml as json arrays."""

    def emit(self, value: AstJson) -> None:
        self.elements.append(value)

    def result(self) -> AstJson:
        node = AstJsonArray(elements=AstChildren(self.elements))
        return set_location(node, node.elements[0], node.elements[-1])


def collect_json_string(
    value: str,
    location: SourceLocation,
    end_location: SourceLocation,
) -> AstJson:
    """Collect yaml string as json node."""
    return set_location(AstJsonValue(value=value), location, end_location)


class NbtCompoundCollector(BaseYamlObjectCollector[AstNbtCompoundEntry, AstNbt]):
    """Collect yaml as nbt compounds."""

    def emit(
        self,
        key: str,
        value: AstNbt,
        location: SourceLocation,
        end_location: SourceLocation,
    ) -> None:
        key_node = AstNbtCompoundKey(value=key)
        key_node = set_location(key_node, location, end_location)
        entry_node = AstNbtCompoundEntry(key=key_node, value=value)
        self.entries.append(set_location(entry_node, key_node, value))

    def result(self) -> AstNbt:
        node = AstNbtCompound(entries=AstChildren(self.entries))
        return set_location(node, node.entries[0], node.entries[-1])


class NbtListCollector(BaseYamlArrayCollector[AstNbt]):
    """Collect yaml as nbt lists."""

    def emit(self, value: AstNbt) -> None:
        self.elements.append(value)

    def result(self) -> AstNbt:
        node = AstNbtList(elements=AstChildren(self.elements))
        return set_location(node, node.elements[0], node.elements[-1])


def collect_nbt_string(
    value: str,
    location: SourceLocation,
    end_location: SourceLocation,
) -> AstNbt:
    """Collect yaml string as nbt node."""
    return set_location(AstNbtValue(value=String(value)), location, end_location)


@dataclass
class NestedYamlParser:
    """Parser for nested yaml."""

    parser: Parser
    original_parser: Parser

    object_collector: BaseYamlObjectCollector[Any, Any]
    array_collector: BaseYamlArrayCollector[Any]
    string_collector: Callable[[str, SourceLocation, SourceLocation], Any]

    quote_helper: QuoteHelper = field(default_factory=JsonQuoteHelper)

    def __call__(self, stream: TokenStream) -> Any:
        nested_yaml = stream.data.get("nested_yaml")

        if not nested_yaml:
            return self.original_parser(stream)

        elif nested_yaml == "init":
            with stream.syntax(colon=r":"):
                if stream.get("colon"):
                    with stream.provide(nested_yaml="active"):
                        return self.parser(stream)
                else:
                    with stream.reset("nested_yaml"):
                        return self.original_parser(stream)

        with (
            stream.intercept("newline"),
            stream.syntax(
                colon=r":",
                dash=r"\-",
                key=rf"[a-zA-Z0-9._+-]+|{YAML_STRING_PATTERN}",
            ),
        ):
            if consume_line_continuation(stream):
                return self.parse_yaml(stream)

        with (
            stream.ignore("newline"),
            stream.syntax(string=YAML_STRING_PATTERN),
        ):
            if token := stream.get("string"):
                return self.string_collector(
                    self.quote_helper.unquote_string(token),
                    token.location,
                    token.end_location,
                )
            else:
                return self.original_parser(stream)

    def parse_yaml(self, stream: TokenStream) -> Any:
        if dash := stream.get("dash"):
            with self.array_collector.start():
                while True:
                    with stream.provide(line_indentation=dash.location.colno - 1):
                        with stream.checkpoint() as commit:
                            self.array_collector.emit(self.parse_yaml(stream))
                            commit()
                        if commit.rollback:
                            self.array_collector.emit(self.parser(stream))

                    if consume_line_continuation(stream):
                        dash = stream.expect("dash")
                    else:
                        return self.array_collector.result()

        key = stream.expect("key")
        stream.expect("colon")

        with self.object_collector.start():
            while True:
                with stream.provide(line_indentation=key.location.colno - 1):
                    self.object_collector.emit(
                        key=self.quote_helper.unquote_string(key),
                        value=self.parser(stream),
                        location=key.location,
                        end_location=key.end_location,
                    )

                if consume_line_continuation(stream):
                    key = stream.expect("key")
                    stream.expect("colon")
                else:
                    return self.object_collector.result()


@dataclass
class NestedYamlContextProvider:
    """Parser that enables nested yaml."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.provide(nested_yaml="init"):
            return self.parser(stream)



__all__ = [
    "Serializer",
    "FormattingOptions",
]


import json
import re
from dataclasses import dataclass, field
from typing import Any, Iterable, Iterator, List, Literal, Union

from beet.core.utils import required_field
from pydantic import BaseModel

from .ast import (
    AstBlock,
    AstBlockState,
    AstBool,
    AstChildren,
    AstCommand,
    AstCoordinate,
    AstItemComponent,
    AstItemPredicate,
    AstItemPredicateAlternatives,
    AstItemPredicateTestComponent,
    AstItemPredicateTestPredicate,
    AstItemRemovedDefaultComponent,
    AstItemStack,
    AstJson,
    AstLiteral,
    AstMacroLine,
    AstMacroLineText,
    AstMacroLineVariable,
    AstMessage,
    AstNbtBool,
    AstNbtByteArray,
    AstNbtCompound,
    AstNbtCompoundKey,
    AstNbtIntArray,
    AstNbtList,
    AstNbtLongArray,
    AstNbtPath,
    AstNbtPathKey,
    AstNbtPathSubscript,
    AstNbtValue,
    AstNode,
    AstNumber,
    AstParticle,
    AstParticleParameters,
    AstRange,
    AstResourceLocation,
    AstRoot,
    AstSelector,
    AstSelectorAdvancementMatch,
    AstSelectorAdvancementPredicateMatch,
    AstSelectorAdvancements,
    AstSelectorArgument,
    AstSelectorScoreMatch,
    AstSelectorScores,
    AstString,
    AstTime,
    AstUUID,
    AstVector2,
    AstVector3,
)
from .database import CompilationDatabase
from .dispatch import Visitor, rule
from .spec import CommandSpec
from .utils import NbtQuoteHelper, QuoteHelper, number_to_string

REGEX_COMMENTS = re.compile(r"^(?:(\s*#.*)|.+)", re.MULTILINE)
UNQUOTED_COMPOUND_KEY = re.compile(r"^[a-zA-Z0-9._+-]+$")


class FormattingOptions(BaseModel):
    """Formatting options."""

    layout: Literal["dense", "preserve"] = "dense"
    cmd_compact: bool = False
    nbt_compact: bool = False
    json_compact: bool = False
    json_ensure_ascii: bool = True
    json_allow_nan: bool = True
    json_sort_keys: bool = False


@dataclass
class Serializer(Visitor):
    spec: CommandSpec = required_field()
    database: CompilationDatabase = field(default_factory=CompilationDatabase)

    formatting: FormattingOptions = field(default_factory=FormattingOptions)

    regex_comments: "re.Pattern[str]" = field(default=REGEX_COMMENTS)
    quote_helper: QuoteHelper = field(
        default_factory=lambda: QuoteHelper(
            escape_sequences={
                r"\\": "\\",
                r"\f": "\f",
                r"\n": "\n",
                r"\r": "\r",
                r"\t": "\t",
            }
        )
    )
    nbt_quote_helper: QuoteHelper = field(default_factory=NbtQuoteHelper)

    def __call__(self, node: AstNode, **kwargs: Any) -> str:
        result: List[str] = []

        previous_formatting = self.formatting if kwargs else None
        if previous_formatting:
            options = previous_formatting.dict()
            options.update(kwargs)
            self.formatting = FormattingOptions.model_validate(options)

        try:
            self.invoke(node, result)
        finally:
            if previous_formatting:
                self.formatting = previous_formatting

        return "".join(result)

    def collection(
        self,
        nodes: Iterable[AstNode],
        delimitters: str,
        result: List[str],
    ) -> Iterator[AstNode]:
        """Helper for serializing collections."""
        result.append(delimitters[0])
        comma = "," if self.formatting.cmd_compact else ", "
        sep = ""
        for node in nodes:
            result.append(sep)
            sep = comma
            yield node
        result.append(delimitters[-1])

    def key_value(self, node: Any, sep: str, result: List[str]) -> Iterator[AstNode]:
        """Helper for serializing key-value pairs."""
        yield node.key
        result.append(sep)
        yield node.value

    @rule(AstNode)
    def unserializable(self, node: AstNode, result: List[str]):
        raise ValueError(f"Couldn't serialize {type(node)!r} node.")

    @rule(AstVector2)
    @rule(AstVector3)
    @rule(AstParticleParameters)
    def aggregate(self, node: AstNode, result: List[str]):
        sep = ""
        for child in node:
            result.append(sep)
            sep = " "
            yield child

    @rule(AstRoot)
    def root(self, node: AstRoot, result: List[str]):
        pos = 0
        source = (
            self.formatting.layout == "preserve"
            and self.database[self.database.current].source
        )

        for command in node.commands:
            if source:
                end_pos = max(command.location.pos, pos + 1)
                if pos > -1:
                    if fill := source[pos:end_pos]:
                        result.append(self.regex_comments.sub(r"\1", fill))
                try:
                    pos = source.index("\n", end_pos) + 1
                except ValueError:
                    pos = -1

            yield command
            result.append("\n")

        if source and pos > -1:
            if fill := source[pos:]:
                result.append(self.regex_comments.sub(r"\1", fill))

    @rule(AstCommand)
    def command(self, node: AstCommand, result: List[str]):
        prototype = self.spec.prototypes[node.identifier]
        argument_index = 0

        sep = ""
        for token in prototype.signature:
            result.append(sep)
            sep = " "

            if isinstance(token, str):
                result.append(token)
            else:
                yield node.arguments[argument_index]
                argument_index += 1

    @rule(AstLiteral)
    def literal(self, node: AstLiteral, result: List[str]):
        result.append(node.value)

    @rule(AstString)
    def string(self, node: AstString, result: List[str]):
        result.append(self.quote_helper.quote_string(node.value))

    @rule(AstBool)
    def bool(self, node: AstBool, result: List[str]):
        result.append(str(node.value).lower())

    @rule(AstNumber)
    def number(self, node: AstNumber, result: List[str]):
        result.append(number_to_string(node.value))

    @rule(AstUUID)
    def uuid(self, node: AstUUID, result: List[str]):
        result.append(str(node.value))

    @rule(AstCoordinate)
    def coordinate(self, node: AstCoordinate, result: List[str]):
        if node.type == "relative":
            result.append("~")
            if node.value == 0:
                return
        elif node.type == "local":
            result.append("^")
            if node.value == 0:
                return
        result.append(number_to_string(node.value))

    @rule(AstJson)
    def json(self, node: AstJson, result: List[str]):
        result.append(
            json.dumps(
                node.evaluate(),
                separators=(",", ":") if self.formatting.json_compact else None,
                ensure_ascii=self.formatting.json_ensure_ascii,
                allow_nan=self.formatting.json_allow_nan,
                sort_keys=self.formatting.json_sort_keys,
            )
        )

    @rule(AstNbtValue)
    def nbt_value(self, node: AstNbtValue, result: List[str]):
        if isinstance(node.value, str):
            result.append(self.nbt_quote_helper.quote_string(node.value))
        else:
            result.append(node.evaluate().snbt(compact=self.formatting.nbt_compact))

    @rule(AstNbtBool)
    def nbt_bool(self, node: AstNbtBool, result: List[str]):
        result.append("true" if node.value else "false")

    @rule(AstNbtList)
    def nbt_list(self, node: AstNbtList, result: List[str]):
        result.append("[")
        comma = "," if self.formatting.nbt_compact else ", "
        sep = ""
        for element in node.elements:
            result.append(sep)
            sep = comma
            yield element
        result.append("]")

    @rule(AstNbtCompoundKey)
    def nbt_compound_key(self, node: AstNbtCompoundKey, result: List[str]):
        if UNQUOTED_COMPOUND_KEY.match(node.value):
            result.append(node.value)
        else:
            result.append(self.nbt_quote_helper.quote_string(node.value))

    @rule(AstNbtCompound)
    def nbt_compound(self, node: AstNbtCompound, result: List[str]):
        result.append("{")
        comma, colon = (",", ":") if self.formatting.nbt_compact else (", ", ": ")
        sep = ""
        for entry in node.entries:
            result.append(sep)
            sep = comma
            yield entry.key
            result.append(colon)
            yield entry.value
        result.append("}")

    @rule(AstNbtByteArray)
    @rule(AstNbtIntArray)
    @rule(AstNbtLongArray)
    def nbt_array(
        self,
        node: Union[AstNbtByteArray, AstNbtIntArray, AstNbtLongArray],
        result: List[str],
    ):
        result.append("[")
        result.append(node.array_prefix)
        semicolon = ";" if self.formatting.nbt_compact else "; "
        result.append(semicolon)
        comma = "," if self.formatting.nbt_compact else ", "
        sep = ""
        for element in node.elements:
            result.append(sep)
            sep = comma
            yield element
        result.append("]")

    @rule(AstResourceLocation)
    def resource_location(self, node: AstResourceLocation, result: List[str]):
        result.append(node.get_value())

    @rule(AstBlockState)
    def block_state(self, node: AstBlockState, result: List[str]):
        yield from self.key_value(node, "=", result)

    @rule(AstBlock)
    def block(self, node: AstBlock, result: List[str]):
        yield node.identifier
        if node.block_states:
            yield from self.collection(node.block_states, "[]", result)
        if node.data_tags:
            yield node.data_tags

    @rule(AstItemComponent)
    def item_component(self, node: AstItemComponent, result: List[str]):
        yield from self.key_value(node, "=", result)

    @rule(AstItemRemovedDefaultComponent)
    def item_removed_default_component(
        self, node: AstItemRemovedDefaultComponent, result: List[str]
    ):
        result.append("!")
        yield node.key

    @rule(AstItemPredicateTestComponent)
    @rule(AstItemPredicateTestPredicate)
    def item_test(
        self,
        node: Union[AstItemPredicateTestComponent, AstItemPredicateTestPredicate],
        result: List[str],
    ):
        if node.inverted:
            result.append("!")

        yield node.key

        if node.value:
            result.append(
                "~" if isinstance(node, AstItemPredicateTestPredicate) else "="
            )
            yield node.value

    @rule(AstItemPredicateAlternatives)
    def item_alternatives(self, node: AstItemPredicateAlternatives, result: List[str]):
        pipe = "|" if self.formatting.cmd_compact else " | "
        sep = ""
        for test_node in node.alternatives:
            result.append(sep)
            sep = pipe
            yield test_node

    @rule(AstItemStack)
    @rule(AstItemPredicate)
    def item(self, node: Union[AstItemStack, AstItemPredicate], result: List[str]):
        yield node.identifier
        if node.arguments:
            yield from self.collection(node.arguments, "[]", result)
        if node.data_tags:
            yield node.data_tags

    @rule(AstRange)
    def range(self, node: AstRange, result: List[str]):
        if node.exact:
            result.append(number_to_string(node.value))
        else:
            if node.min is not None:
                result.append(number_to_string(node.min))
            result.append("..")
            if node.max is not None:
                result.append(number_to_string(node.max))

    @rule(AstTime)
    def time(self, node: AstTime, result: List[str]):
        result.append(number_to_string(node.value))

        if node.unit == "day":
            result.append("d")
        elif node.unit == "second":
            result.append("s")

    @rule(AstSelectorScoreMatch)
    @rule(AstSelectorAdvancementPredicateMatch)
    def selector_argument_key_value(self, node: AstNode, result: List[str]):
        yield from self.key_value(node, "=", result)

    @rule(AstSelectorScores)
    def selector_scores(self, node: AstSelectorScores, result: List[str]):
        yield from self.collection(node.scores, "{}", result)

    @rule(AstSelectorAdvancementMatch)
    def selector_advancement_match(
        self,
        node: AstSelectorAdvancementMatch,
        result: List[str],
    ):
        yield node.key
        result.append("=")
        if isinstance(node.value, AstChildren):
            yield from self.collection(node.value, "{}", result)
        else:
            yield node.value

    @rule(AstSelectorAdvancements)
    def selector_advancements(self, node: AstSelectorAdvancements, result: List[str]):
        yield from self.collection(node.advancements, "{}", result)

    @rule(AstSelectorArgument)
    def selector_argument(self, node: AstSelectorArgument, result: List[str]):
        yield node.key
        result.append("=")
        if node.inverted:
            result.append("!")
        if node.value:
            yield node.value

    @rule(AstSelector)
    def selector(self, node: AstSelector, result: List[str]):
        result.append("@")
        result.append(node.variable)
        if node.arguments:
            yield from self.collection(node.arguments, "[]", result)

    @rule(AstMessage)
    def message(self, node: AstMessage, result: List[str]):
        yield from node.fragments

    @rule(AstNbtPathKey)
    def nbt_path_key(self, node: AstNbtPathKey, result: List[str]):
        value = self.quote_helper.quote_string(node.value)
        if "." in value and not value.startswith(('"', "'")):
            value = f'"{value}"'
        result.append(value)

    @rule(AstNbtPathSubscript)
    def nbt_path_subscript(self, node: AstNbtPathSubscript, result: List[str]):
        result.append("[")
        if node.index:
            yield node.index
        result.append("]")

    @rule(AstNbtPath)
    def nbt_path(self, node: AstNbtPath, result: List[str]):
        sep = ""
        for component in node.components:
            if isinstance(component, AstNbtPathKey):
                result.append(sep)
            sep = "."
            yield component

    @rule(AstParticle)
    def particle(self, node: AstParticle, result: List[str]):
        yield node.name
        if node.parameters:
            if not isinstance(node.parameters, AstNbtCompound):
                result.append(" ")
            yield node.parameters

    @rule(AstMacroLine)
    def macro_line(self, node: AstMacroLine, result: List[str]):
        result.append("$")
        yield from node.arguments

    @rule(AstMacroLineText)
    def macro_line_text(self, node: AstMacroLineText, result: List[str]):
        result.append(node.value)

    @rule(AstMacroLineVariable)
    def macro_line_variable(self, node: AstMacroLineVariable, result: List[str]):
        result.append("$(")
        result.append(node.value)
        result.append(")")



"""Plugin for handling nested resources."""

__all__ = [
    "NestedResources",
    "NestedResourcesTransformer",
    "AstNestedText",
    "parse_nested_text",
]


from dataclasses import dataclass, replace
from textwrap import dedent
from typing import Any, Dict, List, Type, Union, cast

from beet import (
    Context,
    DataModelBase,
    DataPack,
    Generator,
    JsonFileBase,
    NamespaceFile,
    ResourcePack,
    TagFile,
    TextFileBase,
)
from beet.core.utils import JsonDict, required_field, snake_case
from tokenstream import InvalidSyntax, Token, TokenStream, set_location

from mecha import (
    AstChildren,
    AstCommand,
    AstJson,
    AstNode,
    AstResourceLocation,
    AstRoot,
    CompilationDatabase,
    Diagnostic,
    Mecha,
    MultilineParser,
    MutatingReducer,
    consume_line_continuation,
    delegate,
    rule,
)
from mecha.contrib.json_files import (
    AstAppendJsonContent,
    AstJsonContent,
    AstMergeJsonContent,
    AstPrependJsonContent,
    JsonFileCompilation,
)


def beet_default(ctx: Context):
    ctx.require("mecha.contrib.nesting")
    ctx.inject(NestedResources).prepare(ctx.generate, ctx.inject(JsonFileCompilation))


class NestedResources:
    mc: Mecha
    text_resources: Dict[str, Type[NamespaceFile]]
    json_resources: Dict[str, Type[NamespaceFile]]

    def __init__(
        self,
        arg: Union[Mecha, Context],
        *packs: Union[ResourcePack, DataPack],
    ):
        if isinstance(arg, Context):
            packs += arg.packs
            self.mc = arg.inject(Mecha)
        else:
            self.mc = arg

        should_disambiguate = {
            name
            for name, tree in self.mc.spec.tree.get_all_literals()
            if any(tree.get_all_arguments())
        }

        self.text_resources = {
            (
                f"{file_type.snake_name}_file"
                if file_type.snake_name in should_disambiguate
                else file_type.snake_name
            ): file_type
            for pack in packs
            for file_type in pack.get_file_types(extend=TextFileBase)
        }

        self.json_resources = {
            name: self.text_resources.pop(name)
            for name, file_type in list(self.text_resources.items())
            if issubclass(file_type, DataModelBase)
        }

    def prepare(self, generate: Generator, json_file_compilation: JsonFileCompilation):
        text_commands = {
            name: self.create_command_tree_fragment("mecha:nested_text")
            for name in self.text_resources
        }

        json_commands = {
            name: self.create_command_tree_fragment("mecha:nested_json")
            for name in self.json_resources
        }

        tag_commands = {
            name: self.create_command_tree_fragment("mecha:nested_json")
            for name, file_type in self.json_resources.items()
            if issubclass(file_type, TagFile)
        }

        self.mc.spec.add_commands(
            {
                "type": "root",
                "children": {
                    **text_commands,
                    **json_commands,
                    "merge": {"type": "literal", "children": json_commands},
                    "append": {"type": "literal", "children": tag_commands},
                    "prepend": {"type": "literal", "children": tag_commands},
                },
            }
        )

        self.mc.spec.parsers["nested_text"] = parse_nested_text
        self.mc.spec.parsers["command:argument:mecha:nested_text"] = MultilineParser(
            delegate("nested_text")
        )

        self.mc.spec.parsers["nested_json"] = delegate("json")
        self.mc.spec.parsers["command:argument:mecha:nested_json"] = MultilineParser(
            delegate("nested_json")
        )

        self.mc.transform.extend(
            NestedResourcesTransformer(
                generate=generate,
                database=self.mc.database,
                nested_resource_identifiers={
                    f"{prefix}{name}:name:content": file_type
                    for name, file_type in self.text_resources.items()
                    | self.json_resources.items()
                    for prefix in [""]
                    + ["merge:"] * issubclass(file_type, DataModelBase)
                    + ["append:", "prepend:"] * issubclass(file_type, TagFile)
                },
                json_file_compilation=json_file_compilation,
            )
        )

    @staticmethod
    def create_command_tree_fragment(parser: str) -> JsonDict:
        return {
            "type": "literal",
            "children": {
                "name": {
                    "type": "argument",
                    "parser": "minecraft:resource_location",
                    "children": {
                        "content": {
                            "type": "argument",
                            "parser": parser,
                            "executable": True,
                        }
                    },
                }
            },
        }


@dataclass(frozen=True, slots=True)
class AstNestedText(AstNode):
    """Ast nested text node."""

    value: str = required_field()


def parse_nested_text(stream: TokenStream) -> AstNestedText:
    with stream.syntax(colon=r":"):
        colon = stream.expect("colon")

    with stream.intercept("newline"), stream.syntax(text=r"[^\s].*", comment=None):
        if not consume_line_continuation(stream):
            exc = InvalidSyntax("Expected non-empty block.")
            raise set_location(exc, colon)

        start_pos = stream.current.location.pos
        level = stream.indentation[-2]

        lines: List[Token] = []

        while True:
            lines.append(stream.expect("text"))

            with stream.provide(multiline=True, line_indentation=level):
                if not consume_line_continuation(stream):
                    break

    node = AstNestedText(
        value=dedent(stream.source[start_pos : lines[-1].end_location.pos] + "\n")
    )
    return set_location(node, lines[0], lines[-1])


@dataclass
class NestedResourcesTransformer(MutatingReducer):
    """Transformer that handles nested resources."""

    generate: Generator = required_field()
    database: CompilationDatabase = required_field()

    nested_resource_identifiers: Dict[str, Type[NamespaceFile]] = required_field()
    json_file_compilation: JsonFileCompilation = required_field()

    @rule(AstRoot)
    def nested_resources(self, node: AstRoot):
        changed = False
        commands: List[AstCommand] = []

        for command in node.commands:
            if file_type := self.nested_resource_identifiers.get(command.identifier):
                name, content = command.arguments

                if isinstance(name, AstResourceLocation) and isinstance(
                    content, (AstJson, AstNestedText)
                ):
                    full_name = name.get_canonical_value()

                    if isinstance(content, AstJson) and (
                        cast(Type[JsonFileBase[Any]], file_type)
                        in self.json_file_compilation.file_types
                    ):
                        if command.identifier.startswith("merge:"):
                            content_type = AstMergeJsonContent
                        elif command.identifier.startswith("append:"):
                            content_type = AstAppendJsonContent
                        elif command.identifier.startswith("prepend:"):
                            content_type = AstPrependJsonContent
                        else:
                            content_type = AstJsonContent

                        root = content_type.create_root_node(content)

                        file_instance = file_type(
                            original=self.database.current.original
                        )
                        target = cast(  # pyright: ignore[reportInvalidCast]
                            JsonFileBase[Any],
                            self.generate(full_name, default=file_instance),
                        )

                        compilation_unit = self.database.get(target)
                        if not compilation_unit:
                            compilation_unit = replace(
                                self.database[self.database.current],
                                ast=(
                                    root
                                    if target is file_instance
                                    else AstJsonContent.create_root_node(
                                        AstJson.from_value(target.data)
                                    )
                                ),
                                resource_location=full_name,
                                no_index=True,
                                pack=(
                                    self.generate.data
                                    if file_type
                                    in self.generate.data.resolve_scope_map().values()
                                    else self.generate.assets
                                ),
                            )
                            self.database[target] = compilation_unit
                            self.database.enqueue(target, self.database.step + 1)

                        if compilation_unit.ast:
                            if content_type is AstJsonContent:
                                for command in compilation_unit.ast.commands:
                                    if isinstance(command, AstJsonContent):
                                        if command != root.commands[0]:
                                            d = Diagnostic(
                                                level="error",
                                                message=f'Redefinition of {snake_case(file_type.__name__)} "{full_name}" doesn\'t match existing file.',
                                            )
                                            yield set_location(d, name)
                                        break
                            else:
                                compilation_unit.ast = replace(
                                    compilation_unit.ast,
                                    commands=AstChildren(
                                        compilation_unit.ast.commands + root.commands
                                    ),
                                )
                        else:
                            compilation_unit.ast = root

                        changed = True
                        continue

                    file_instance = file_type(
                        (
                            content.evaluate()
                            if isinstance(content, AstJson)
                            else content.value
                        ),
                        original=self.database.current.original,
                    )

                    if command.identifier.startswith("merge:"):
                        self.generate(full_name, merge=file_instance)
                        changed = True
                        continue

                    target = self.generate(full_name, default=file_instance)
                    if target is not file_instance:
                        if command.identifier.startswith("append:"):
                            target.append(file_instance)  # pyright: ignore[reportAttributeAccessIssue]
                        elif command.identifier.startswith("prepend:"):
                            target.prepend(file_instance)  # pyright: ignore[reportAttributeAccessIssue]
                        elif (
                            target.ensure_deserialized()
                            != file_instance.ensure_deserialized()
                        ):
                            d = Diagnostic(
                                level="error",
                                message=f'Redefinition of {snake_case(file_type.__name__)} "{full_name}" doesn\'t match existing file.',
                            )
                            yield set_location(d, name)

                    changed = True
                    continue

            commands.append(command)

        if changed:
            return replace(node, commands=AstChildren(commands))

        return node

    @rule(AstCommand, identifier="execute:run:subcommand")
    def execute_nested_resource(self, node: AstCommand):
        if (
            isinstance(subcommand := node.arguments[0], AstCommand)
            and subcommand.identifier in self.nested_resource_identifiers
        ):
            d = Diagnostic("error", "Nested resource not allowed behind execute.")
            yield set_location(d, subcommand, subcommand.arguments[0])
        return node



__all__ = [
    "validate",
    "mecha",
    "main",
]


import logging
import os
import zipfile
from typing import Optional, Tuple

import click
from beet import LATEST_MINECRAFT_VERSION, Context, DataPack, Function, run_beet
from beet.toolchain.cli import BeetCommand, LogHandler, error_handler, message_fence

from mecha import __version__

from .api import Mecha


def validate(ctx: Context):
    """Plugin for running mecha on the provided sources."""
    mc = ctx.inject(Mecha)

    for path in ctx.meta["source"]:
        path = ctx.directory / path

        if zipfile.is_zipfile(path) or (path / "data").is_dir():
            mc.compile(DataPack(path=path), report=mc.diagnostics)

        elif path.is_dir():
            for filename in sorted(path.glob("**/*.mcfunction")):
                mc.compile(Function(source_path=filename), report=mc.diagnostics)

        elif path.is_file():
            mc.compile(Function(source_path=path), report=mc.diagnostics)


@click.command(
    cls=BeetCommand,
    context_settings={"help_option_names": ("-h", "--help")},
)
@click.argument("source", nargs=-1, type=click.Path(exists=True))
@click.option(
    "-m",
    "--minecraft",
    metavar="VERSION",
    default=LATEST_MINECRAFT_VERSION,
    help="Minecraft version.",
)
@click.option(
    "-l",
    "--log",
    metavar="LEVEL",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        case_sensitive=False,
    ),
    default="WARNING",
    help="Configure output verbosity.",
)
@click.option(
    "-s",
    "--stats",
    is_flag=True,
    help="Collect statistics.",
)
@click.option(
    "-j",
    "--json",
    metavar="FILENAME",
    help="Output json.",
)
@click.version_option(
    __version__,
    "-v",
    "--version",
    message=click.style("%(prog)s", fg="red")
    + click.style(" v%(version)s", fg="green"),
)
@error_handler(should_exit=True)
def mecha(
    source: Tuple[str, ...],
    minecraft: str,
    log: str,
    stats: bool,
    json: Optional[str],
):
    """Validate data packs and .mcfunction files."""
    if stats and log != "DEBUG":
        log = "INFO"

    logger = logging.getLogger()
    logger.setLevel(log)
    logger.addHandler(LogHandler())

    config = {
        "minecraft": minecraft,
        "require": stats * ["mecha.contrib.statistics"],
        "pipeline": ["mecha.cli.validate"],
        "meta": {
            "source": source,
            "mecha": {
                "readonly": True,
                "cache": False,
            },
            "statistics": {
                "output": os.path.abspath(json) if json else None,
            },
        },
    }

    with message_fence(f"Validating with mecha v{__version__}"):
        if source:
            with run_beet(config):
                pass
        else:
            logger.warning("No path provided.", extra={"prefix": "mecha"})
            logger.warning("Use --help to see usage.", extra={"prefix": "mecha"})


def main():
    """Invoke the command-line entrypoint."""
    mecha(prog_name="mecha")



__all__ = [
    "Mecha",
    "MechaOptions",
    "AstCacheBackend",
    "FORMATTING_PRESETS",
]


import csv
import logging
import os
import pickle
import sys
from contextlib import contextmanager
from dataclasses import InitVar, dataclass
from glob import glob
from io import BufferedReader, BufferedWriter
from pathlib import Path
from time import perf_counter_ns
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from beet import (
    LATEST_MINECRAFT_VERSION,
    Cache,
    Context,
    DataPack,
    Function,
    ListOption,
    PackageablePath,
    ResourcePack,
    TextFileBase,
)
from beet.core.utils import (
    FileSystemPath,
    JsonDict,
    VersionNumber,
    extra_field,
    import_from_string,
    resolve_within,
)
from pydantic import BaseModel, field_validator
from tokenstream import InvalidSyntax, Preprocessor, TokenStream, set_location

from .ast import AstLiteral, AstNode, AstRoot
from .config import CommandTree
from .database import (
    CompilationDatabase,
    CompilationUnit,
    CompilationUnitProvider,
    FileTypeCompilationUnitProvider,
)
from .diagnostic import (
    Diagnostic,
    DiagnosticCollection,
    DiagnosticError,
    DiagnosticErrorSummary,
)
from .dispatch import Dispatcher, MutatingReducer, Reducer
from .parse import delegate, get_parsers
from .preprocess import wrap_backslash_continuation
from .serialize import FormattingOptions, Serializer
from .spec import CommandSpec

AstNodeType = TypeVar("AstNodeType", bound=AstNode)
TextFileType = TypeVar("TextFileType", bound=TextFileBase[Any])
PackType = TypeVar("PackType", bound=Union[ResourcePack, DataPack])


logger = logging.getLogger("mecha")


@dataclass
class AstCacheBackend:
    """Backend for the ast cache."""

    version: str = import_from_string("mecha.__version__")

    def load_data(self, f: BufferedReader) -> JsonDict:
        """Load the pickled data."""
        data = pickle.load(f)
        if data["mecha"] != self.version or data["python"] != sys.version:
            raise ValueError("Version mismatch.")
        return data

    def dump_data(self, data: JsonDict, f: BufferedWriter):
        """Dump the pickled data."""
        data["mecha"] = self.version
        data["python"] = sys.version
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    def load(self, f: BufferedReader) -> AstRoot:
        """Load the ast."""
        return self.load_data(f)["ast"]

    def dump(self, node: AstRoot, f: BufferedWriter):
        """Dump the ast."""
        self.dump_data({"ast": node}, f)


FORMATTING_PRESETS: Dict[str, JsonDict] = {
    "minify": {
        "layout": "dense",
        "cmd_compact": True,
        "nbt_compact": True,
        "json_compact": True,
    },
    "dense": {
        "layout": "dense",
    },
    "preserve": {
        "layout": "preserve",
    },
}


class MechaOptions(BaseModel):
    """Mecha options."""

    version: str = ""
    commands: Optional[ListOption[PackageablePath]] = None
    multiline: bool = False
    formatting: FormattingOptions = FormattingOptions()
    readonly: Optional[bool] = None
    match: Optional[List[str]] = None
    rules: Dict[str, Literal["ignore", "info", "warn", "error"]] = {}
    cache: bool = True
    output_perf: Optional[FileSystemPath] = None

    @field_validator("formatting", mode="before")
    @classmethod
    def formatting_preset(cls, value: Any):
        if isinstance(value, str):
            assert value in FORMATTING_PRESETS, "invalid formatting preset"
            return FORMATTING_PRESETS[value]
        return value


@dataclass
class Mecha:
    """Class exposing the command api."""

    ctx: InitVar[Optional[Context]] = None
    version: InitVar[VersionNumber] = LATEST_MINECRAFT_VERSION
    multiline: InitVar[bool] = False
    formatting: InitVar[Optional[FormattingOptions]] = None
    readonly: bool = False
    match: Optional[List[str]] = None

    directory: Path = extra_field(init=False)
    cache: Optional[Cache] = extra_field(default=None)
    cache_backend: AstCacheBackend = extra_field(default_factory=AstCacheBackend)
    output_perf: Optional[FileSystemPath] = extra_field(default=None)
    perf_report: Optional[List[Tuple[str, str, int, List[float]]]] = extra_field(
        default=None
    )

    spec: CommandSpec = extra_field(default=None)

    function_provider: CompilationUnitProvider = extra_field(init=False)
    providers: List[CompilationUnitProvider] = extra_field(init=False)

    preprocessor: Preprocessor = extra_field(default=wrap_backslash_continuation)

    lint: Dispatcher[AstRoot] = extra_field(init=False)
    transform: Dispatcher[AstRoot] = extra_field(init=False)
    optimize: Dispatcher[AstRoot] = extra_field(init=False)
    check: Dispatcher[AstRoot] = extra_field(init=False)

    steps: List[Dispatcher[AstRoot]] = extra_field(default_factory=list)

    serialize: Serializer = extra_field(init=False)

    database: CompilationDatabase = extra_field(default_factory=CompilationDatabase)
    diagnostics: DiagnosticCollection = extra_field(
        default_factory=DiagnosticCollection
    )

    def __post_init__(
        self,
        ctx: Optional[Context],
        version: VersionNumber,
        multiline: bool,
        formatting: Optional[FormattingOptions],
    ):
        if ctx:
            opts = ctx.validate("mecha", MechaOptions)
            version = opts.version or ctx.minecraft_version
            multiline = opts.multiline
            formatting = opts.formatting

            if opts.readonly is not None:
                self.readonly = opts.readonly
            if opts.match is not None:
                self.match = opts.match

            self.directory = ctx.directory
            self.database.configure(generate=ctx.generate, directory=self.directory)

            self.lint = Reducer(levels=opts.rules)
            self.transform = MutatingReducer(levels=opts.rules)
            self.optimize = MutatingReducer(levels=opts.rules)
            self.check = Reducer(levels=opts.rules)

            if not self.cache and opts.cache:
                self.cache = ctx.cache["mecha"]

            if opts.output_perf:
                self.output_perf = ctx.directory / opts.output_perf

            if opts.commands is not None:
                commands = [
                    p
                    for entry in opts.commands.entries()
                    for p in (
                        [ctx.cache["commands"].download(entry.http_url)]
                        if entry.http_url
                        else sorted(glob(str(ctx.directory / entry)))
                    )
                ]
                self.spec = CommandSpec(
                    multiline=multiline,
                    tree=CommandTree.load_from(
                        *commands,
                        version=version,
                        patch_only=True,
                    ),
                    parsers=get_parsers(version),
                )

            ctx.require(self.finalize)

        else:
            self.directory = Path.cwd()
            self.database.configure(directory=self.directory)

            self.lint = Reducer()
            self.transform = MutatingReducer()
            self.optimize = MutatingReducer()
            self.check = Reducer()

        self.steps.append(self.lint)
        self.steps.append(self.transform)
        self.steps.append(self.optimize)
        self.steps.append(self.check)

        if self.perf_report is None and self.output_perf:
            self.perf_report = []

        if not self.spec:
            self.spec = CommandSpec(
                multiline=multiline,
                tree=CommandTree.load_from(version=version),
                parsers=get_parsers(version),
            )

        self.function_provider = FileTypeCompilationUnitProvider([Function])
        self.providers = [self.function_provider]

        self.serialize = Serializer(
            spec=self.spec,
            database=self.database,
            formatting=formatting or FormattingOptions(),
        )

    @contextmanager
    def prepare_token_stream(
        self,
        stream: TokenStream,
        multiline: Optional[bool] = None,
    ) -> Iterator[TokenStream]:
        """Prepare the token stream for parsing."""
        with (
            stream.reset(*stream.data),
            stream.provide(
                spec=self.spec,
                multiline=self.spec.multiline if multiline is None else multiline,
            ),
        ):
            with stream.reset_syntax(comment=r"#.*$", literal=AstLiteral.regex.pattern):
                with stream.indent(skip=["comment"]), stream.ignore("indent", "dedent"):
                    with stream.intercept("newline", "eof"):
                        yield stream

    @overload
    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
        preprocessor: Optional[Preprocessor] = None,
    ) -> AstRoot: ...

    @overload
    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        type: Type[AstNodeType],
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
        preprocessor: Optional[Preprocessor] = None,
    ) -> AstNodeType: ...

    @overload
    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        using: str,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
        preprocessor: Optional[Preprocessor] = None,
    ) -> Any: ...

    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        type: Optional[Type[AstNode]] = None,
        using: Optional[str] = None,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
        preprocessor: Optional[Preprocessor] = None,
    ) -> Any:
        """Parse the given source into an AST."""
        if using:
            parser = using
        else:
            if not type:
                type = AstRoot
            if not type.parser:
                raise TypeError(f"No parser directly associated with {type}.")
            parser = type.parser

        if isinstance(source, (list, str)):
            source = Function(source)

        if not filename:
            filename = resolve_within(source.source_path, self.directory)

        cache_miss = None

        if self.cache and filename:
            ast_path = self.cache.get_path(f"{self.directory / filename}-ast")

            if not self.cache.has_changed(self.directory / filename):
                try:
                    with ast_path.open("rb") as f:
                        ast = self.cache_backend.load(f)
                        logger.debug('Load cached ast for file "%s".', filename)
                        return ast
                except Exception:
                    pass

            cache_miss = ast_path

        stream = TokenStream(
            source.text,
            preprocessor=preprocessor or self.preprocessor,
        )

        try:
            with self.prepare_token_stream(stream, multiline=multiline):
                with stream.provide(**provide or {}):
                    ast = delegate(parser, stream)
        except InvalidSyntax as exc:
            if self.cache and filename and cache_miss:
                self.cache.invalidate_changes(self.directory / filename)
            d = Diagnostic(
                level="error",
                message=str(exc),
                notes=exc.notes,
                hint=resource_location,
                filename=str(filename) if filename else None,
                file=source,
            )
            raise DiagnosticError(DiagnosticCollection([set_location(d, exc)])) from exc
        else:
            if self.cache and filename and cache_miss:
                try:
                    with cache_miss.open("wb") as f:
                        self.cache_backend.dump(ast, f)
                        logger.debug('Update cached ast for file "%s".', filename)
                except Exception:
                    pass
            return ast

    @overload
    def compile(
        self,
        source: PackType,
        *,
        match: Optional[List[str]] = None,
        multiline: Optional[bool] = None,
        formatting: Optional[JsonDict] = None,
        readonly: Optional[bool] = None,
        initial_step: int = 0,
        report: Optional[DiagnosticCollection] = None,
    ) -> PackType: ...

    @overload
    def compile(
        self,
        *,
        together: Iterable[Union[ResourcePack, DataPack]],
        match: Optional[List[str]] = None,
        multiline: Optional[bool] = None,
        formatting: Optional[JsonDict] = None,
        readonly: Optional[bool] = None,
        initial_step: int = 0,
        report: Optional[DiagnosticCollection] = None,
    ) -> None: ...

    @overload
    def compile(
        self,
        source: TextFileType,
        *,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        within: Optional[Union[ResourcePack, DataPack]] = None,
        multiline: Optional[bool] = None,
        formatting: Optional[JsonDict] = None,
        readonly: Optional[bool] = None,
        initial_step: int = 0,
        report: Optional[DiagnosticCollection] = None,
    ) -> TextFileType: ...

    @overload
    def compile(
        self,
        source: Union[List[str], str, AstRoot],
        *,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        within: Optional[Union[ResourcePack, DataPack]] = None,
        multiline: Optional[bool] = None,
        formatting: Optional[JsonDict] = None,
        readonly: Optional[bool] = None,
        initial_step: int = 0,
        report: Optional[DiagnosticCollection] = None,
    ) -> Function: ...

    def compile(
        self,
        source: Optional[
            Union[
                Union[ResourcePack, DataPack],
                TextFileBase[Any],
                List[str],
                str,
                AstRoot,
            ]
        ] = None,
        *,
        together: Optional[Iterable[Union[ResourcePack, DataPack]]] = None,
        match: Optional[List[str]] = None,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        within: Optional[Union[ResourcePack, DataPack]] = None,
        multiline: Optional[bool] = None,
        formatting: Optional[JsonDict] = None,
        readonly: Optional[bool] = None,
        initial_step: int = 0,
        report: Optional[DiagnosticCollection] = None,
    ) -> Optional[Union[Union[ResourcePack, DataPack], TextFileBase[Any]]]:
        """Apply all compilation steps."""
        self.database.setup_compilation()

        if formatting is None:
            formatting = {}
        if readonly is None:
            readonly = self.readonly

        result = None

        if isinstance(source, (ResourcePack, DataPack)):
            result = source
            together = [source]

        if together is not None:
            if match is None:
                match = self.match

            for pack in together:
                for provider in self.providers:
                    for file_instance, compilation_unit in provider(pack, match):
                        self.database[file_instance] = compilation_unit
                        self.database.enqueue(file_instance)
        else:
            if isinstance(source, (list, str)):
                source = Function(source)

            if isinstance(source, TextFileBase):
                result = source
                if not filename:
                    filename = resolve_within(source.source_path, self.directory)
            else:
                result = Function()

            compilation_unit = CompilationUnit(
                resource_location=resource_location,
                filename=str(filename) if filename else None,
                pack=within,
            )

            if isinstance(source, AstRoot):
                compilation_unit.ast = source

            self.database[result] = compilation_unit
            self.database.enqueue(result)

        for step, file_instance in self.database.process_queue():
            compilation_unit = self.database[file_instance]
            start_time = perf_counter_ns()

            if step < 0:
                if compilation_unit.ast:
                    self.database.enqueue(file_instance, initial_step)
                    continue
                try:
                    compilation_unit.source = file_instance.text
                    compilation_unit.ast = self.parse(
                        file_instance,
                        filename=compilation_unit.filename,
                        resource_location=compilation_unit.resource_location,
                        multiline=multiline,
                    )
                    self.database.enqueue(file_instance, initial_step)
                except DiagnosticError as exc:
                    compilation_unit.diagnostics.extend(exc.diagnostics)

            elif step < len(self.steps):
                if not compilation_unit.ast:
                    continue
                with self.steps[step].use_diagnostics(compilation_unit.diagnostics):
                    if ast := self.steps[step](compilation_unit.ast):
                        if not compilation_unit.diagnostics.error:
                            compilation_unit.ast = ast
                            self.database.enqueue(
                                key=file_instance,
                                step=step + 1,
                                priority=compilation_unit.priority,
                            )
            else:
                if readonly:
                    continue
                if not compilation_unit.ast:
                    continue
                with self.serialize.use_diagnostics(compilation_unit.diagnostics):
                    file_instance.text = self.serialize(
                        compilation_unit.ast, **formatting
                    )

            compilation_unit.perf[step] = (perf_counter_ns() - start_time) * 1e-06

        sorted_source_files = sorted(
            self.database.session,
            key=lambda f: self.database[f].resource_location or "<unknown>",
        )

        if self.perf_report is not None:
            for file_instance in sorted_source_files:
                compilation_unit = self.database[file_instance]
                self.perf_report.append(
                    (
                        compilation_unit.filename or "",
                        compilation_unit.resource_location or "",
                        len(list(compilation_unit.diagnostics.get_all_errors())),
                        [
                            compilation_unit.perf.get(i, float("nan"))
                            for i in range(-1, len(self.steps) + 1)
                        ],
                    )
                )

        diagnostics = DiagnosticCollection(
            [
                exc
                for file_instance in sorted_source_files
                for exc in self.database[file_instance].diagnostics.exceptions
            ]
        )

        if report:
            report.extend(diagnostics)
        else:
            if errors := list(diagnostics.get_all_errors()):
                raise DiagnosticError(DiagnosticCollection(errors))

        return result

    def log_reported_diagnostics(self):
        """Log reported diagnostics."""
        for diagnostic in self.diagnostics.exceptions:
            message = diagnostic.format_message()

            if diagnostic.file:
                if source := self.database[diagnostic.file].source:
                    if code := diagnostic.format_code(source):
                        message += f"\n{code}"

            if diagnostic.notes:
                message += f"\n{diagnostic.format_notes()}"

            extra = {"annotate": diagnostic.format_location()}

            if diagnostic.level == "error":
                logger.error("%s", message, extra=extra)
            elif diagnostic.level == "warn":
                logger.warning("%s", message, extra=extra)
            elif diagnostic.level == "info":
                logger.info("%s", message, extra=extra)

    def format_perf(self) -> List[List[str]]:
        """Format perf report."""
        step_headers = [
            (
                "Lint"
                if step is self.lint
                else (
                    "Transform"
                    if step is self.transform
                    else (
                        "Optimize"
                        if step is self.optimize
                        else "Check"
                        if step is self.check
                        else repr(step)
                    )
                )
            )
            for step in self.steps
        ]

        return [
            ["Filename", "Resource location", "Errors", "Parse"]
            + step_headers
            + ["Serialize"]
        ] + [
            [filename, resource_location, str(errors)] + [f"{t:.3f}" for t in steps]
            for filename, resource_location, errors, steps in self.perf_report or []
        ]

    def finalize(self, ctx: Context):
        """Plugin that logs diagnostics and raises an exception if there are errors."""
        try:
            yield
        except:
            raise
        else:
            if errors := list(self.diagnostics.get_all_errors()):
                raise DiagnosticErrorSummary(DiagnosticCollection(errors))
        finally:
            if path := self.output_perf:
                logger.info('Output perf "%s".', os.path.relpath(path, self.directory))
                with open(path, "w") as f:
                    csv.writer(f).writerows(self.format_perf())
            self.log_reported_diagnostics()



__all__ = [
    "CompilationDatabase",
    "CompilationUnit",
    "CompilationUnitProvider",
    "FileTypeCompilationUnitProvider",
]


from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from heapq import heappop, heappush
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from beet import (
    Container,
    DataPack,
    Generator,
    NamespaceFile,
    ResourcePack,
    TextFile,
    TextFileBase,
)
from beet.core.utils import FileSystemPath, extra_field, resolve_within

from .ast import AstRoot
from .diagnostic import DiagnosticCollection


@dataclass
class CompilationUnit:
    """Information associated with the compilation of a specific function file."""

    ast: Optional[AstRoot] = None
    source: Optional[str] = None
    filename: Optional[str] = None
    resource_location: Optional[str] = None
    no_index: bool = False
    pack: Optional[Union[ResourcePack, DataPack]] = None
    priority: int = 0

    diagnostics: DiagnosticCollection = extra_field(init=False)
    perf: Dict[int, float] = extra_field(init=False)

    def __post_init__(self):
        self.diagnostics = DiagnosticCollection(
            filename=self.filename,
            hint=self.resource_location,
        )
        self.perf = {}


class CompilationDatabase(Container[TextFileBase[Any], CompilationUnit]):
    """Compilation database."""

    indices: DefaultDict[
        Optional[Union[ResourcePack, DataPack]], Dict[str, TextFileBase[Any]]
    ]
    session: Set[TextFileBase[Any]]
    queue: List[Tuple[int, int, str, int, TextFileBase[Any]]]
    step: int
    current: TextFileBase[Any]
    count: int

    generate: Optional[Generator]
    directory: Optional[FileSystemPath]

    def __init__(self):
        super().__init__()
        self.indices = defaultdict(dict)
        self.session = set()
        self.queue = []
        self.step = -1
        self.current = TextFile()
        self.count = 0

        self.generate = None
        self.directory = None

    def configure(
        self,
        generate: Optional[Generator] = None,
        directory: Optional[FileSystemPath] = None,
    ):
        self.generate = generate
        self.directory = directory
        self[self.current] = CompilationUnit()

    @property
    def index(self) -> Dict[str, TextFileBase[Any]]:
        return self.indices[self[self.current].pack]

    def process(
        self,
        key: TextFileBase[Any],
        value: CompilationUnit,
    ) -> CompilationUnit:
        if not value.filename:
            value.filename = resolve_within(key.source_path, self.directory)
        if not value.diagnostics.filename and value.filename:
            value.diagnostics.filename = value.filename

        value.diagnostics.file = key

        if not value.no_index:
            pack_index = self.indices[value.pack]
            for index in [value.filename, value.resource_location]:
                if index:
                    pack_index[index] = key

        return value

    def __delitem__(self, key: TextFileBase[Any]):
        pack_index = self.indices[self[key].pack]
        for index in [self[key].filename, self[key].resource_location]:
            if index:
                del pack_index[index]
        super().__delitem__(key)

    def setup_compilation(self):
        """Setup database for compilation."""
        self.session.clear()
        self.queue.clear()

    def enqueue(self, key: TextFileBase[Any], step: int = -1, priority: int = 0):
        """Enqueue a file and schedule it to be processed with the given step."""
        self.session.add(key)
        self.count += 1
        heappush(
            self.queue,
            (
                step,
                priority,
                self[key].resource_location or "<unknown>",
                self.count,
                key,
            ),
        )

    def dequeue(self) -> Tuple[int, TextFileBase[Any]]:
        """Dequeue the next file that needs to be processed."""
        step, _, _, _, key = heappop(self.queue)
        return step, key

    def process_queue(self) -> Iterator[Tuple[int, TextFileBase[Any]]]:
        """Yield database entries from the queue."""
        while self.queue:
            self.step, self.current = self.dequeue()
            with self.process_context(self.current):
                yield self.step, self.current

    @contextmanager
    def process_context(self, file_instance: TextFileBase[Any]):
        if self.generate is not None:
            compilation_unit = self[file_instance]
            if compilation_unit.pack is not None:
                with self.generate.overlays[compilation_unit.pack.overlay_name].push():
                    yield
                    return
        yield


class CompilationUnitProvider(Protocol):
    """Provide source files for compilation."""

    def __call__(
        self,
        pack: Union[ResourcePack, DataPack],
        match: Optional[List[str]] = None,
    ) -> Iterable[Tuple[TextFileBase[Any], CompilationUnit]]: ...


@dataclass
class FileTypeCompilationUnitProvider:
    """Provide source files based on their type."""

    file_types: Sequence[Type[NamespaceFile]]
    no_index: bool = False

    def __call__(
        self,
        pack: Union[ResourcePack, DataPack],
        match: Optional[List[str]] = None,
    ) -> Iterable[Tuple[TextFileBase[Any], CompilationUnit]]:
        packs = [pack]
        if pack.overlay_parent is None:
            packs.extend(pack.overlays.values())

        for file_type in self.file_types:
            if not issubclass(file_type, TextFileBase):
                continue

            for pack in packs:
                for resource_location in pack[file_type].match(*match or ["*"]):
                    file_instance = pack[file_type][resource_location]

                    yield (
                        file_instance,
                        CompilationUnit(
                            resource_location=resource_location,
                            no_index=self.no_index,
                            pack=pack,
                        ),
                    )
